@@ -15,6 +15,12 @@
 #include <stdio.h>
 #include <errno.h>
 
+#ifdef HAVE_STDBOOL_H
+#include <stdbool.h>
+#else
+#include "missing/stdbool.h"
+#endif
+
 #ifdef USE_ELF
 
 #include <fcntl.h>
@@ -94,6 +100,10 @@ void *alloca();
 #define PATH_MAX 4096
 #endif
 
+#ifndef SHF_COMPRESSED /* compatibility with glibc < 2.22 */
+#define SHF_COMPRESSED 0
+#endif
+
 int kprintf(const char *fmt, ...);
 
 typedef struct {
@@ -427,22 +437,25 @@ fill_lines(int num_traces, void **traces, int check_debuglink,
 	   obj_info_t **objp, line_info_t *lines, int offset);
 
 static void
-append_obj(obj_info_t **objp) {
+append_obj(obj_info_t **objp)
+{
     obj_info_t *newobj = calloc(1, sizeof(obj_info_t));
     if (*objp) (*objp)->next = newobj;
     *objp = newobj;
 }
 
 static void
-follow_debuglink(char *debuglink, int num_traces, void **traces,
+follow_debuglink(const char *debuglink, int num_traces, void **traces,
 		 obj_info_t **objp, line_info_t *lines, int offset)
 {
     /* Ideally we should check 4 paths to follow gnu_debuglink,
        but we handle only one case for now as this format is used
        by some linux distributions. See GDB's info for detail. */
     static const char global_debug_dir[] = "/usr/lib/debug";
-    char *p, *subdir;
+    const size_t global_debug_dir_len = sizeof(global_debug_dir) - 1;
+    char *p;
     obj_info_t *o1 = *objp, *o2;
+    size_t len;
 
     p = strrchr(binary_filename, '/');
     if (!p) {
@@ -450,11 +463,13 @@ follow_debuglink(char *debuglink, int num_traces, void **traces,
     }
     p[1] = '\0';
 
-    subdir = (char *)alloca(strlen(binary_filename) + 1);
-    strcpy(subdir, binary_filename);
-    strcpy(binary_filename, global_debug_dir);
-    strlcat(binary_filename, subdir, PATH_MAX);
-    strlcat(binary_filename, debuglink, PATH_MAX);
+    len = strlen(binary_filename);
+    if (len >= PATH_MAX - global_debug_dir_len)
+	len = PATH_MAX - global_debug_dir_len - 1;
+    memmove(binary_filename + global_debug_dir_len, binary_filename, len);
+    memcpy(binary_filename, global_debug_dir, global_debug_dir_len);
+    len += global_debug_dir_len;
+    strlcpy(binary_filename + len, debuglink, PATH_MAX - len);
 
     append_obj(objp);
     o2 = *objp;
@@ -481,6 +496,7 @@ fill_lines(int num_traces, void **traces, int check_debuglink,
     ElfW(Shdr) *dynsym_shdr = NULL, *dynstr_shdr = NULL;
     obj_info_t *obj = *objp;
     uintptr_t dladdr_fbase = 0;
+    bool compressed_p = false;
 
     fd = open(binary_filename, O_RDONLY);
     if (fd < 0) {
@@ -550,6 +566,9 @@ fill_lines(int num_traces, void **traces, int check_debuglink,
 	    break;
 	  case SHT_PROGBITS:
 	    if (!strcmp(section_name, ".debug_line")) {
+		if (shdr[i].sh_flags & SHF_COMPRESSED) {
+		    compressed_p = true;
+		}
 		debug_line_shdr = shdr + i;
 	    }
 	    else if (!strcmp(section_name, ".gnu_debuglink")) {
@@ -627,7 +646,8 @@ fill_lines(int num_traces, void **traces, int check_debuglink,
 	goto finish;
     }
 
-    if (parse_debug_line(num_traces, traces,
+    if (!compressed_p &&
+	    parse_debug_line(num_traces, traces,
 			 file + debug_line_shdr->sh_offset,
 			 debug_line_shdr->sh_size,
 			 obj, lines, offset))
@@ -649,7 +669,7 @@ fail:
  * it is NUL terminated.
  */
 #if defined(__linux__)
-ssize_t
+static ssize_t
 main_exe_path(void)
 {
 # define PROC_SELF_EXE "/proc/self/exe"
@@ -658,7 +678,7 @@ main_exe_path(void)
     return len;
 }
 #elif defined(__FreeBSD__)
-ssize_t
+static ssize_t
 main_exe_path(void)
 {
     int mib[4] = {CTL_KERN, KERN_PROC, KERN_PROC_PATHNAME, -1};
@@ -726,7 +746,7 @@ rb_dump_backtrace_with_lines(int num_traces, void **traces)
 	    path = info.dli_fname;
 	    obj->path = path;
 	    lines[i].path = path;
-	    strcpy(binary_filename, path);
+	    strlcpy(binary_filename, path, PATH_MAX);
 	    if (fill_lines(num_traces, traces, 1, &obj, lines, i) == (uintptr_t)-1)
 		break;
 	}
@@ -743,7 +763,7 @@ next_line:
 	    kprintf("[0x%lx]\n", addr);
 	}
 	else if (!line->saddr || !line->sname) {
-	    kprintf("%s [0x%lx]\n", line->path, addr);
+	    kprintf("%s(0x%lx) [0x%lx]\n", line->path, addr-line->base_addr, addr);
 	}
 	else if (line->line <= 0) {
 	    kprintf("%s(%s+0x%lx) [0x%lx]\n", line->path, line->sname,
