@@ -253,6 +253,22 @@ f_finite_p(VALUE x)
     return RTEST(rb_funcallv(x, id_finite_p, 0, 0));
 }
 
+VALUE rb_flo_is_infinite_p(VALUE num);
+inline static VALUE
+f_infinite_p(VALUE x)
+{
+    if (RB_INTEGER_TYPE_P(x)) {
+        return Qnil;
+    }
+    else if (RB_FLOAT_TYPE_P(x)) {
+	return rb_flo_is_infinite_p(x);
+    }
+    else if (RB_TYPE_P(x, T_RATIONAL)) {
+        return Qnil;
+    }
+    return rb_funcallv(x, id_infinite_p, 0, 0);
+}
+
 inline static int
 f_kind_of_p(VALUE x, VALUE c)
 {
@@ -282,6 +298,7 @@ nucomp_s_new_internal(VALUE klass, VALUE real, VALUE imag)
 
     RCOMPLEX_SET_REAL(obj, real);
     RCOMPLEX_SET_IMAG(obj, imag);
+    OBJ_FREEZE_RAW(obj);
 
     return (VALUE)obj;
 }
@@ -1336,8 +1353,8 @@ nucomp_inspect(VALUE self)
  * call-seq:
  *    cmp.finite?  ->  true or false
  *
- * Returns +true+ if +cmp+'s magnitude is finite number,
- * oterwise returns +false+.
+ * Returns +true+ if +cmp+'s magnitude is a finite number,
+ * otherwise returns +false+.
  */
 static VALUE
 rb_complex_finite_p(VALUE self)
@@ -1352,7 +1369,7 @@ rb_complex_finite_p(VALUE self)
 
 /*
  * call-seq:
- *    cmp.infinite?  ->  nil or 1 or -1
+ *    cmp.infinite?  ->  nil or 1
  *
  * Returns values corresponding to the value of +cmp+'s magnitude:
  *
@@ -1367,21 +1384,12 @@ rb_complex_finite_p(VALUE self)
 static VALUE
 rb_complex_infinite_p(VALUE self)
 {
-    VALUE magnitude = nucomp_abs(self);
+    get_dat1(self);
 
-    if (FINITE_TYPE_P(magnitude)) {
+    if (NIL_P(f_infinite_p(dat->real)) && NIL_P(f_infinite_p(dat->imag))) {
 	return Qnil;
     }
-    if (RB_FLOAT_TYPE_P(magnitude)) {
-	const double f = RFLOAT_VALUE(magnitude);
-	if (isinf(f)) {
-	    return INT2FIX(f < 0 ? -1 : 1);
-	}
-	return Qnil;
-    }
-    else {
-	return rb_funcall(magnitude, id_infinite_p, 0);
-    }
+    return ONE;
 }
 
 /* :nodoc: */
@@ -1399,6 +1407,7 @@ nucomp_loader(VALUE self, VALUE a)
 
     RCOMPLEX_SET_REAL(dat, rb_ivar_get(a, id_i_real));
     RCOMPLEX_SET_IMAG(dat, rb_ivar_get(a, id_i_imag));
+    OBJ_FREEZE_RAW(self);
 
     return self;
 }
@@ -1456,20 +1465,6 @@ rb_Complex(VALUE x, VALUE y)
     return nucomp_s_convert(2, a, rb_cComplex);
 }
 
-VALUE
-rb_complex_set_real(VALUE cmp, VALUE r)
-{
-    RCOMPLEX_SET_REAL(cmp, r);
-    return cmp;
-}
-
-VALUE
-rb_complex_set_imag(VALUE cmp, VALUE i)
-{
-    RCOMPLEX_SET_IMAG(cmp, i);
-    return cmp;
-}
-
 VALUE
 rb_complex_abs(VALUE cmp)
 {
