@@ -6,20 +6,16 @@
 $extension = nil
 $extstatic = nil
 $force_static = nil
-$install = nil
 $destdir = nil
 $dryrun = false
-$clean = nil
 $nodynamic = nil
-$extinit = nil
 $extobjs = []
 $extflags = ""
 $extlibs = nil
 $extpath = nil
-$ignore = nil
 $message = nil
 $command_output = nil
-$configure_only = false
+$subconfigure = false
 
 $progname = $0
 alias $PROGRAM_NAME $0
@@ -106,7 +102,7 @@ def extract_makefile(makefile, keep = true)
   srcs = Dir[File.join($srcdir, "*.{#{SRC_EXT.join(%q{,})}}")].map {|fn| File.basename(fn)}.sort
   if !srcs.empty?
     old_srcs = m[/^ORIG_SRCS[ \t]*=[ \t](.*)/, 1] or return false
-    old_srcs.split.sort == srcs or return false
+    (old_srcs.split - srcs).empty? or return false
   end
   $target = target
   $extconf_h = m[/^RUBY_EXTCONF_H[ \t]*=[ \t]*(\S+)/, 1]
@@ -131,12 +127,7 @@ def extract_makefile(makefile, keep = true)
   true
 end
 
-def extmake(target, basedir = (maybestatic = 'ext'))
-  unless $configure_only || verbose?
-    print "#{$message} #{target}\n"
-    $stdout.flush
-  end
-
+def extmake(target, basedir = 'ext', maybestatic = true)
   FileUtils.mkpath target unless File.directory?(target)
   begin
     # don't build if parent library isn't build
@@ -170,7 +161,7 @@ def extmake(target, basedir = (maybestatic = 'ext'))
     static = $static
     $static = nil if noinstall = File.fnmatch?("-*", target)
     ok = parent && File.exist?(makefile)
-    if parent && !$ignore
+    if parent
       rbconfig0 = RbConfig::CONFIG
       mkconfig0 = CONFIG
       rbconfig = {
@@ -206,14 +197,12 @@ def extmake(target, basedir = (maybestatic = 'ext'))
 	    [conf, "#{$srcdir}/depend"].any? {|f| modified?(f, [t])})
         then
 	  ok = false
-          if $configure_only
-            if verbose?
-              print "#{conf}\n" if conf
-            else
-              print "#{$message} #{target}\n"
-            end
-            $stdout.flush
+          if verbose?
+            print "#{conf}\n" if conf
+          else
+            print "#{$message} #{target}\n"
           end
+          $stdout.flush
           init_mkmf
 	  Logging::logfile 'mkmf.log'
 	  rm_f makefile
@@ -233,7 +222,6 @@ def extmake(target, basedir = (maybestatic = 'ext'))
       rescue SystemExit
 	# ignore
       rescue => error
-        lineno = error.backtrace_locations[0].lineno
         ok = false
       ensure
 	rm_f "conftest*"
@@ -249,38 +237,27 @@ def extmake(target, basedir = (maybestatic = 'ext'))
 
       return true if !error and target.start_with?("-")
 
-      if parent
-        message = "Failed to configure #{target}. It will not be installed."
-      else
-        message = "Skipped to configure #{target}. Its parent is not configured."
+      message = nil
+      if error
+        loc = error.backtrace_locations[0]
+        message = "#{loc.absolute_path}:#{loc.lineno}: #{error.message}"
+        if Logging.log_opened?
+          Logging::message("#{message}\n\t#{error.backtrace.join("\n\t")}\n")
+        end
       end
-      if Logging.log_opened?
-        Logging::message(error.to_s) if error
-        Logging::message(message)
-      end
-      message = error.message if error
 
-      return parent ? [conf, lineno||0, message] : true
+      return [parent, message]
     end
     args = $mflags
     unless $destdir.to_s.empty? or $mflags.defined?("DESTDIR")
       args += ["DESTDIR=" + relative_from($destdir, "../"+prefix)]
     end
     if $static and ok and !$objs.empty? and !noinstall
-      args += ["static"] unless $clean
+      args += ["static"]
       $extlist.push [(maybestatic ? $static : false), target, $target, $preload]
     end
     FileUtils.rm_f(old_cleanfiles - $distcleanfiles - $cleanfiles)
     FileUtils.rm_f(old_objs - $objs)
-    unless $configure_only or system($make, *args)
-      $ignore or $continue or return false
-    end
-    if $clean
-      FileUtils.rm_f("mkmf.log")
-      if $clean != true
-	FileUtils.rm_f([makefile, $extconf_h || "extconf.h"])
-      end
-    end
     if $static
       $extflags ||= ""
       $extlibs ||= []
@@ -376,6 +353,7 @@ def parse_args()
     $optparser.warn(e)
     abort $optparser.to_s
   end
+  $command_output or abort "--command-output option is mandatory"
 
   $destdir ||= ''
 
@@ -411,18 +389,10 @@ parse_args()
 if target = ARGV.shift and /^[a-z-]+$/ =~ target
   $mflags.push(target)
   case target
-  when /^(dist|real)?(clean)$/
-    target = $2
-    $ignore ||= true
-    $clean = $1 ? $1[0] : true
-  when /^install\b/
-    $install = true
-    $ignore ||= true
-    $mflags.unshift("INSTALL_PROG=install -c -p -m 0755",
-                    "INSTALL_DATA=install -c -p -m 0644",
-                    "MAKEDIRS=mkdir -p") if $dryrun
+  when /^(dist|real)?(clean)$/, /^install\b/
+    abort "#{target} is obsolete"
   when /configure/
-    $configure_only = true
+    $subconfigure = !ARGV.empty?
   end
 end
 unless $message
@@ -441,12 +411,15 @@ elsif sep = config_string('BUILD_FILE_SEPARATOR')
 else
   $ruby = '$(topdir)/miniruby' + EXEEXT
 end
-$ruby << " -I'$(topdir)'"
+$ruby = [$ruby]
+$ruby << "-I'$(topdir)'"
 unless CROSS_COMPILING
-  $ruby << " -I'$(top_srcdir)/lib'"
-  $ruby << " -I'$(extout)/$(arch)' -I'$(extout)/common'" if $extout
+  $ruby << "-I'$(top_srcdir)/lib'"
+  $ruby << "-I'$(extout)/$(arch)'" << "-I'$(extout)/common'" if $extout
   ENV["RUBYLIB"] = "-"
 end
+topruby = $ruby
+$ruby = topruby.join(' ')
 $mflags << "ruby=#$ruby"
 
 MTIMES = [__FILE__, 'rbconfig.rb', srcdir+'/lib/mkmf.rb'].collect {|f| File.mtime(f)}
@@ -486,7 +459,13 @@ for dir in ["ext", File::join($top_srcdir, "ext")]
   end
 end unless $extstatic
 
-ext_prefix = "#{$top_srcdir}/ext"
+@gemname = nil
+if ARGV[0]
+  ext_prefix, exts = ARGV.shift.split('/', 2)
+  $extension = [exts] if exts
+  @gemname = exts if ext_prefix == 'gems'
+end
+ext_prefix = "#{$top_srcdir}/#{ext_prefix || 'ext'}"
 exts = $static_ext.sort_by {|t, i| i}.collect {|t, i| t}
 default_exclude_exts =
   case
@@ -497,6 +476,7 @@ default_exclude_exts =
   else
     %w'*win32*'
   end
+mandatory_exts = {}
 withes, withouts = [["--with", nil], ["--without", default_exclude_exts]].collect {|w, d|
   if !(w = %w[-extensions -ext].collect {|o|arg_config(w+o)}).any?
     d ? proc {|c1| d.any?(&c1)} : proc {true}
@@ -504,13 +484,15 @@ withes, withouts = [["--with", nil], ["--without", default_exclude_exts]].collec
     proc {true}
   else
     w = w.collect {|o| o.split(/,/)}.flatten
-    w.collect! {|o| o == '+' ? d : o}.flatten! if d
+    w.collect! {|o| o == '+' ? d : o}.flatten!
     proc {|c1| w.any?(&c1)}
   end
 }
 cond = proc {|ext, *|
-  cond1 = proc {|n| File.fnmatch(n, ext)}
-  withes.call(cond1) and !withouts.call(cond1)
+  withes.call(proc {|n|
+                !n or (mandatory_exts[ext] = true if File.fnmatch(n, ext))
+              }) and
+    !withouts.call(proc {|n| File.fnmatch(n, ext)})
 }
 ($extension || %w[*]).each do |e|
   e = e.sub(/\A(?:\.\/)+/, '')
@@ -525,63 +507,22 @@ cond = proc {|ext, *|
     exts.delete_if {|d| File.fnmatch?("-*", d)}
   end
 end
-
-if $extout
-  extout = RbConfig.expand("#{$extout}", RbConfig::CONFIG.merge("topdir"=>$topdir))
-  unless $ignore
-    FileUtils.mkpath("#{extout}/gems")
-  end
-end
-
-FileUtils.makedirs('gems')
-ext_prefix = "#$top_srcdir/gems"
-gems = Dir.glob(File.join(ext_prefix, ($extension || ''), '**/extconf.rb')).collect {|d|
-  d = File.dirname(d)
-  d.slice!(0, ext_prefix.length + 1)
-  d
-}.find_all {|ext|
-  with_config(ext, &cond)
-}.sort
+ext_prefix = File.basename(ext_prefix)
 
 extend Module.new {
   def timestamp_file(name, target_prefix = nil)
+    if @gemname and name == '$(TARGET_SO_DIR)'
+      name = "$(arch)/gems/#{@gemname}#{target_prefix}"
+    end
     super.sub(%r[/\.extout\.(?:-\.)?], '/.')
   end
 
   def configuration(srcdir)
     super << "EXTSO #{['=', $extso].join(' ')}\n"
   end
-}
 
-dir = Dir.pwd
-FileUtils::makedirs('ext')
-Dir::chdir('ext')
-
-hdrdir = $hdrdir
-$hdrdir = ($top_srcdir = relative_from(srcdir, $topdir = "..")) + "/include"
-extso = []
-fails = []
-exts.each do |d|
-  $static = $force_static ? true : $static_ext[d]
-
-  if $ignore or !$nodynamic or $static
-    result = extmake(d) or abort
-    extso |= $extso
-    fails << result unless result == true
-  end
-end
-
-Dir.chdir('..')
-FileUtils::makedirs('gems')
-Dir.chdir('gems')
-extout = $extout
-unless gems.empty?
-  def self.timestamp_file(name, target_prefix = nil)
-    name = "$(arch)/gems/#{@gemname}#{target_prefix}" if name == '$(TARGET_SO_DIR)'
-    super
-  end
-
-  def self.create_makefile(*args, &block)
+  def create_makefile(*args, &block)
+    return super unless @gemname
     super(*args) do |conf|
       conf.find do |s|
         s.sub!(/^(TARGET_SO_DIR *= *)\$\(RUBYARCHDIR\)/) {
@@ -600,19 +541,31 @@ build_complete: $(build_complete)
 $(build_complete): $(TARGET_SO)
 	$(Q) $(TOUCH) $@
 
+clean-so::
+	-$(Q)$(RM) $(build_complete)
 }
       conf
     end
   end
+}
+
+dir = Dir.pwd
+FileUtils::makedirs(ext_prefix)
+Dir::chdir(ext_prefix)
+
+hdrdir = $hdrdir
+$hdrdir = ($top_srcdir = relative_from(srcdir, $topdir = "..")) + "/include"
+extso = []
+fails = []
+exts.each do |d|
+  $static = $force_static ? true : $static_ext[d]
+
+  if !$nodynamic or $static
+    result = extmake(d, ext_prefix, !@gemname) or abort
+    extso |= $extso
+    fails << [d, result] unless result == true
+  end
 end
-gems.each do |d|
-  $extout = extout.dup
-  @gemname = d[%r{\A[^/]+}]
-  extmake(d, 'gems')
-  extso |= $extso
-end
-$extout = extout
-Dir.chdir('../ext')
 
 $top_srcdir = srcdir
 $topdir = "."
@@ -623,32 +576,13 @@ extinit = Struct.new(:c, :o) {
     super("#{src}.c", "#{src}.#{$OBJEXT}")
   end
 }.new("extinit")
-if $ignore
-  FileUtils.rm_f(extinit.to_a) if $clean
-  Dir.chdir ".."
-  if $clean
-    Dir.rmdir('ext') rescue nil
-    if $extout
-      FileUtils.rm_rf([extout+"/common", extout+"/include/ruby", extout+"/rdoc"])
-      FileUtils.rm_rf(extout+"/"+CONFIG["arch"])
-      if $clean != true
-	FileUtils.rm_rf(extout+"/include/"+CONFIG["arch"])
-	FileUtils.rm_f($mflags.defined?("INSTALLED_LIST")||ENV["INSTALLED_LIST"]||".installed.list")
-	Dir.rmdir(extout+"/include") rescue nil
-	Dir.rmdir(extout) rescue nil
-      end
-    end
-  end
-  exit
-end
 
-$extinit ||= ""
 $extobjs ||= []
 $extpath ||= []
 $extflags ||= ""
 $extlibs ||= []
+extinits = []
 unless $extlist.empty?
-  $extinit << "\n" unless $extinit.empty?
   list = $extlist.dup
   built = []
   while e = list.shift
@@ -662,27 +596,11 @@ unless $extlist.empty?
       next
     end
     base = File.basename(feature)
-    $extinit << "    init(Init_#{base}, \"#{feature}.so\");\n"
+    extinits << feature
     $extobjs << format("ext/%s/%s.%s", target, base, $LIBEXT)
     built << target
   end
 
-  src = %{\
-#include "ruby/ruby.h"
-
-#define init(func, name) {	\\
-    extern void func(void);	\\
-    ruby_init_ext(name, func);	\\
-}
-
-void ruby_init_ext(const char *name, void (*init)(void));
-
-void Init_ext(void)\n{\n#$extinit}
-}
-  if !modified?(extinit.c, MTIMES) || IO.read(extinit.c) != src
-    open(extinit.c, "w") {|fe| fe.print src}
-  end
-
   $extpath.delete("$(topdir)")
   $extflags = libpathflag($extpath) << " " << $extflags.strip
   conf = [
@@ -692,7 +610,7 @@ void Init_ext(void)\n{\n#$extinit}
   ].map {|n, v|
     "#{n}=#{v}" if v &&= v[/\S(?:.*\S)?/]
   }.compact
-  puts(*conf)
+  puts(*conf) unless $subconfigure
   $stdout.flush
   $mflags.concat(conf)
   $makeflags.concat(conf)
@@ -716,9 +634,9 @@ $makeflags.uniq!
 
 $mflags.unshift("topdir=#$topdir")
 ENV.delete("RUBYOPT")
-if $configure_only and $command_output
-  exts.map! {|d| "ext/#{d}/."}
-  gems.map! {|d| "gems/#{d}/."}
+exts.map! {|d| "#{ext_prefix}/#{d}/."}
+FileUtils.makedirs(File.dirname($command_output))
+begin
   atomic_write_open($command_output) do |mf|
     mf.puts "V = 0"
     mf.puts "Q1 = $(V:1=)"
@@ -743,12 +661,14 @@ if $configure_only and $command_output
       puts
     end
 
+    mf.macro "ruby", topruby
+    mf.macro "RUBY", ["$(ruby)"]
     mf.macro "extensions", exts
-    mf.macro "gems", gems
     mf.macro "EXTOBJS", $extlist.empty? ? ["dmyext.#{$OBJEXT}"] : ["ext/extinit.#{$OBJEXT}", *$extobjs]
     mf.macro "EXTLIBS", $extlibs
     mf.macro "EXTSO", extso
     mf.macro "EXTLDFLAGS", $extflags.split
+    mf.macro "EXTINITS", extinits
     submakeopts = []
     if enable_config("shared", $enable_shared)
       submakeopts << 'DLDOBJS="$(EXTOBJS) $(EXTENCS)"'
@@ -760,21 +680,23 @@ if $configure_only and $command_output
       submakeopts << 'EXTLIBS="$(EXTLIBS)"'
     end
     submakeopts << 'EXTLDFLAGS="$(EXTLDFLAGS)"'
+    submakeopts << 'EXTINITS="$(EXTINITS)"'
     submakeopts << 'UPDATE_LIBRARIES="$(UPDATE_LIBRARIES)"'
     submakeopts << 'SHOWFLAGS='
     mf.macro "SUBMAKEOPTS", submakeopts
+    mf.macro "NOTE_MESG", %w[$(RUBY) $(top_srcdir)/tool/colorize.rb skip]
+    mf.macro "NOTE_NAME", %w[$(RUBY) $(top_srcdir)/tool/colorize.rb fail]
     mf.puts
     targets = %w[all install static install-so install-rb clean distclean realclean]
     targets.each do |tgt|
       mf.puts "#{tgt}: $(extensions:/.=/#{tgt})"
-      mf.puts "#{tgt}: $(gems:/.=/#{tgt})" unless tgt == 'static'
       mf.puts "#{tgt}: note" unless /clean\z/ =~ tgt
     end
     mf.puts
     mf.puts "clean:\n\t-$(Q)$(RM) ext/extinit.#{$OBJEXT}"
     mf.puts "distclean:\n\t-$(Q)$(RM) ext/extinit.c"
     mf.puts
-    mf.puts "#{rubies.join(' ')}: $(extensions:/.=/#{$force_static ? 'static' : 'all'}) $(gems:/.=/all)"
+    mf.puts "#{rubies.join(' ')}: $(extensions:/.=/#{$force_static ? 'static' : 'all'})"
     submake = "$(Q)$(MAKE) $(MFLAGS) $(SUBMAKEOPTS)"
     mf.puts "all static: #{rubies.join(' ')}\n"
     $extobjs.each do |tgt|
@@ -785,7 +707,7 @@ if $configure_only and $command_output
       mf.puts "#{tgt}:\n\t#{submake} $@"
     end
     mf.puts "libencs:\n\t$(Q)$(MAKE) -f enc.mk V=$(V) $@"
-    mf.puts "ext/extinit.#{$OBJEXT}:\n\t$(Q)$(MAKE) $(MFLAGS) V=$(V) $@" if $static
+    mf.puts "ext/extinit.#{$OBJEXT}:\n\t$(Q)$(MAKE) $(MFLAGS) V=$(V) EXTINITS=\"$(EXTINITS)\" $@" if $static
     mf.puts
     if $gnumake == "yes"
       submake = "$(MAKE) -C $(@D)"
@@ -794,9 +716,8 @@ if $configure_only and $command_output
       config_string("exec") {|str| submake << str << " "}
       submake << "$(MAKE)"
     end
-    gems = exts + gems
     targets.each do |tgt|
-      (tgt == 'static' ? exts : gems).each do |d|
+      exts.each do |d|
         mf.puts "#{d[0..-2]}#{tgt}:\n\t$(Q)#{submake} $(MFLAGS) V=$(V) $(@F)"
       end
     end
@@ -805,43 +726,32 @@ if $configure_only and $command_output
 
     mf.puts "\n""note:\n"
     unless fails.empty?
-      mf.puts %Q<\t@echo "*** Following extensions failed to configure:">
-      fails.each do |d, n, err|
-        d = "#{d}:#{n}:"
-        if err
-          d << " " << err
+      abandon = false
+      mf.puts "note: note-body\n"
+      mf.puts "note-body:: note-header\n"
+      mf.puts "note-header:\n"
+      mf.puts %Q<\t@$(NOTE_MESG) "*** Following extensions are not compiled:">
+      mf.puts "note-body:: note-header\n"
+      fails.each do |ext, (parent, err)|
+        abandon ||= mandatory_exts[ext]
+        mf.puts %Q<\t@$(NOTE_NAME) "#{ext}:">
+        if parent
+          mf.puts %Q<\t@echo "\tCould not be configured. It will not be installed.">
+          err and err.scan(/.+/) do |ee|
+            mf.puts %Q<\t@echo "\t#{ee.gsub(/["`$^]/, '\\\\\\&')}">
+          end
+          mf.puts %Q<\t@echo "\tCheck #{ext_prefix}/#{ext}/mkmf.log for more details.">
+        else
+          mf.puts %Q<\t@echo "\tSkipped because its parent was not configured.">
         end
-        mf.puts %Q<\t@echo "#{d}">
       end
-      mf.puts %Q<\t@echo "*** Fix the problems, then remove these directories and try again if you want.">
-    end
-
-  end
-elsif $command_output
-  message = "making #{rubies.join(', ')}"
-  message = "echo #{message}"
-  $mflags.concat(rubies)
-  $makeflags.concat(rubies)
-  cmd = $makeflags.map {|ss|ss.sub(/.*[$(){};\s].*/, %q['\&'])}.join(' ')
-  open($command_output, 'wb') do |ff|
-    case $command_output
-    when /\.sh\z/
-      ff.puts message, "rm -f \"$0\"; exec \"$@\" #{cmd}"
-    when /\.bat\z/
-      ["@echo off", message, "%* #{cmd}", "del %0 & exit %ERRORLEVEL%"].each do |ss|
-        ff.print ss, "\r\n"
+      mf.puts "note:\n"
+      mf.puts %Q<\t@$(NOTE_MESG) "*** Fix the problems, then remove these directories and try again if you want.">
+      if abandon
+        mf.puts "\t""@exit 1"
       end
-    else
-      ff.puts cmd
     end
-    ff.chmod(0755)
   end
-elsif !$configure_only
-  message = "making #{rubies.join(', ')}"
-  puts message
-  $stdout.flush
-  $mflags.concat(rubies)
-  system($make, *$mflags) or exit($?.exitstatus)
 end
 # :startdoc:
 
