@@ -1,3 +1,4 @@
+# frozen_string_literal: true
 # = net/pop.rb
 #
 # Copyright (c) 1999-2007 Yukihiro Matsumoto.
@@ -13,9 +14,9 @@
 # Ruby Distribute License.
 #
 # NOTE: You can find Japanese version of this document at:
-# http://www.ruby-lang.org/ja/man/html/net_pop.html
+# http://docs.ruby-lang.org/ja/latest/library/net=2fpop.html
 #
-#   $Id: pop.rb 40295 2013-04-14 15:19:46Z nagachika $
+#   $Id: pop.rb 59472 2017-08-02 15:10:42Z kazu $
 #
 # See Net::POP3 for documentation.
 #
@@ -167,8 +168,8 @@ module Net
   #     require 'net/pop'
   #
   #     # Use APOP authentication if $isapop == true
-  #     pop = Net::POP3.APOP($is_apop).new('apop.example.com', 110)
-  #     pop.start(YourAccount', 'YourPassword') do |pop|
+  #     pop = Net::POP3.APOP($isapop).new('apop.example.com', 110)
+  #     pop.start('YourAccount', 'YourPassword') do |pop|
   #       # Rest of the code is the same.
   #     end
   #
@@ -195,7 +196,7 @@ module Net
   class POP3 < Protocol
 
     # svn revision of this library
-    Revision = %q$Revision: 40295 $.split[1]
+    Revision = %q$Revision: 59472 $.split[1]
 
     #
     # Class Parameters
@@ -549,15 +550,15 @@ module Net
         context.set_params(@ssl_params)
         s = OpenSSL::SSL::SSLSocket.new(s, context)
         s.sync_close = true
-        s.connect
+        ssl_socket_connect(s, @open_timeout)
         if context.verify_mode != OpenSSL::SSL::VERIFY_NONE
           s.post_connection_check(@address)
         end
       end
-      @socket = InternetMessageIO.new(s)
+      @socket = InternetMessageIO.new(s,
+                                      read_timeout: @read_timeout,
+                                      debug_output: @debug_output)
       logging "POP session started: #{@address}:#{@port} (#{@apop ? 'APOP' : 'POP'})"
-      @socket.read_timeout = @read_timeout
-      @socket.debug_output = @debug_output
       on_connect
       @command = POP3Command.new(@socket)
       if apop?
@@ -569,7 +570,7 @@ module Net
     ensure
       # Authentication failed, clean up connection.
       unless @started
-        s.close if s and not s.closed?
+        s.close if s
         @socket = nil
         @command = nil
       end
@@ -600,7 +601,7 @@ module Net
     ensure
       @started = false
       @command = nil
-      @socket.close if @socket and not @socket.closed?
+      @socket.close if @socket
       @socket = nil
     end
     private :do_finish
@@ -707,7 +708,7 @@ module Net
       @mails.each {|m| m.uid = uidl[m.number] }
     end
 
-    # deguging output for +msg+
+    # debugging output for +msg+
     def logging(msg)
       @debug_output << msg + "\n" if @debug_output
     end
@@ -770,7 +771,7 @@ module Net
     # === Example without block
     #
     #     POP3.start('pop.example.com', 110,
-    #                'YourAccount, 'YourPassword') do |pop|
+    #                'YourAccount', 'YourPassword') do |pop|
     #       n = 1
     #       pop.mails.each do |popmail|
     #         File.open("inbox/#{n}", 'w') do |f|
@@ -784,7 +785,7 @@ module Net
     # === Example with block
     #
     #     POP3.start('pop.example.com', 110,
-    #                'YourAccount, 'YourPassword') do |pop|
+    #                'YourAccount', 'YourPassword') do |pop|
     #       n = 1
     #       pop.mails.each do |popmail|
     #         File.open("inbox/#{n}", 'w') do |f|
@@ -843,7 +844,7 @@ module Net
     # === Example
     #
     #     POP3.start('pop.example.com', 110,
-    #                'YourAccount, 'YourPassword') do |pop|
+    #                'YourAccount', 'YourPassword') do |pop|
     #       n = 1
     #       pop.mails.each do |popmail|
     #         File.open("inbox/#{n}", 'w') do |f|
@@ -889,7 +890,7 @@ module Net
 
     def initialize(sock)
       @socket = sock
-      @error_occured = false
+      @error_occurred = false
       res = check_response(critical { recv_response() })
       @apop_stamp = res.slice(/<[!-~]+@[!-~]+>/)
     end
@@ -1007,11 +1008,11 @@ module Net
     end
 
     def critical
-      return '+OK dummy ok response' if @error_occured
+      return '+OK dummy ok response' if @error_occurred
       begin
         return yield()
       rescue Exception
-        @error_occured = true
+        @error_occurred = true
         raise
       end
     end
