@@ -1,8 +1,8 @@
 /**********************************************************************
 
-  eval.c -
+  safe.c -
 
-  $Author: akr $
+  $Author: ko1 $
   created at: Tue Sep 23 09:44:32 JST 2008
 
   Copyright (C) 2008 Yukihiro Matsumoto
@@ -12,40 +12,50 @@
 /* safe-level:
    0 - strings from streams/environment/ARGV are tainted (default)
    1 - no dangerous operation by tainted value
-   2 - process/file operations prohibited
-   3 - all generated objects are tainted
-   4 - no global (non-tainted) variable modification/no direct output
 */
 
-#define SAFE_LEVEL_MAX 4
+#define SAFE_LEVEL_MAX RUBY_SAFE_LEVEL_MAX
 
 #include "ruby/ruby.h"
 #include "vm_core.h"
 
 /* $SAFE accessor */
 
+#undef rb_secure
+#undef rb_set_safe_level
+#undef ruby_safe_level_2_warning
+
+int
+ruby_safe_level_2_warning(void)
+{
+    return 2;
+}
+
 int
 rb_safe_level(void)
 {
-    return GET_THREAD()->safe_level;
+    return GET_EC()->safe_level;
 }
 
 void
 rb_set_safe_level_force(int safe)
 {
-    GET_THREAD()->safe_level = safe;
+    GET_EC()->safe_level = safe;
 }
 
 void
 rb_set_safe_level(int level)
 {
-    rb_thread_t *th = GET_THREAD();
+    rb_execution_context_t *ec = GET_EC();
 
-    if (level > th->safe_level) {
+    if (level > ec->safe_level) {
 	if (level > SAFE_LEVEL_MAX) {
-	    level = SAFE_LEVEL_MAX;
+	    rb_raise(rb_eArgError, "$SAFE=2 to 4 are obsolete");
 	}
-	th->safe_level = level;
+	/* block parameters */
+	rb_vm_stack_to_heap(ec);
+
+	ec->safe_level = level;
     }
 }
 
@@ -58,30 +68,36 @@ safe_getter(void)
 static void
 safe_setter(VALUE val)
 {
+    rb_execution_context_t *ec = GET_EC();
+    int current_level = ec->safe_level;
     int level = NUM2INT(val);
-    rb_thread_t *th = GET_THREAD();
 
-    if (level < th->safe_level) {
+    if (level == current_level) {
+	return;
+    }
+    else if (level < current_level) {
 	rb_raise(rb_eSecurityError,
 		 "tried to downgrade safe level from %d to %d",
-		 th->safe_level, level);
+		 current_level, level);
     }
-    if (level == 3) {
-	rb_warning("$SAFE=3 does no sandboxing; you might want to use $SAFE=4");
+    else if (level > SAFE_LEVEL_MAX) {
+	rb_raise(rb_eArgError, "$SAFE=2 to 4 are obsolete");
     }
-    if (level > SAFE_LEVEL_MAX) {
-	level = SAFE_LEVEL_MAX;
-    }
-    th->safe_level = level;
+
+    /* block parameters */
+    rb_vm_stack_to_heap(ec);
+
+    ec->safe_level = level;
 }
 
 void
 rb_secure(int level)
 {
     if (level <= rb_safe_level()) {
-	if (rb_frame_callee()) {
-	    rb_raise(rb_eSecurityError, "Insecure operation `%s' at level %d",
-		     rb_id2name(rb_frame_callee()), rb_safe_level());
+	ID caller_name = rb_frame_callee();
+	if (caller_name) {
+	    rb_raise(rb_eSecurityError, "Insecure operation `%"PRIsVALUE"' at level %d",
+		     rb_id2str(caller_name), rb_safe_level());
 	}
 	else {
 	    rb_raise(rb_eSecurityError, "Insecure operation at level %d",
@@ -93,16 +109,15 @@ rb_secure(int level)
 void
 rb_secure_update(VALUE obj)
 {
-    if (!OBJ_TAINTED(obj))
-	rb_secure(4);
 }
 
 void
 rb_insecure_operation(void)
 {
-    if (rb_frame_callee()) {
-	rb_raise(rb_eSecurityError, "Insecure operation - %s",
-		 rb_id2name(rb_frame_callee()));
+    ID caller_name = rb_frame_callee();
+    if (caller_name) {
+	rb_raise(rb_eSecurityError, "Insecure operation - %"PRIsVALUE,
+		 rb_id2str(caller_name));
     }
     else {
 	rb_raise(rb_eSecurityError, "Insecure operation: -r");
@@ -115,17 +130,6 @@ rb_check_safe_obj(VALUE x)
     if (rb_safe_level() > 0 && OBJ_TAINTED(x)) {
 	rb_insecure_operation();
     }
-    rb_secure(4);
-}
-
-void
-rb_check_safe_str(VALUE x)
-{
-    rb_check_safe_obj(x);
-    if (TYPE(x) != T_STRING) {
-	rb_raise(rb_eTypeError, "wrong argument type %s (expected String)",
-		 rb_obj_classname(x));
-    }
 }
 
 void
