@@ -1,3 +1,4 @@
+# frozen_string_literal: false
 require 'drb/drb'
 require 'drb/extserv'
 
@@ -11,4 +12,5 @@ if __FILE__ == $0
   DRb.start_service('druby://:8473', [1, 2, 'III', 4, "five", 6])
   es = DRb::ExtServ.new(ARGV.shift, ARGV.shift)
   DRb.thread.join
+  es.stop_service if es.alive?
 end
