@@ -1,4 +1,4 @@
-# frozen_string_literal: false
+# frozen_string_literal: true
 require 'minitest/unit'
 require 'pp'
 
@@ -89,11 +89,13 @@ module Test
           }
 
           return e
+        ensure
+          unless e
+            exp = exp.first if exp.size == 1
+
+            flunk(message(msg) {"#{mu_pp(exp)} expected but nothing was raised"})
+          end
         end
-
-        exp = exp.first if exp.size == 1
-
-        flunk(message(msg) {"#{mu_pp(exp)} expected but nothing was raised"})
       end
 
       def assert_raises(*exp, &b)
@@ -131,14 +133,20 @@ module Test
           raise TypeError, "Expected #{expected.inspect} to be a kind of String or Regexp, not #{expected.class}"
         end
 
-        ex = assert_raise(exception, msg || proc {"Exception(#{exception}) with message matches to #{expected.inspect}"}) {yield}
+        ex = m = nil
+        EnvUtil.with_default_internal(expected.encoding) do
+          ex = assert_raise(exception, msg || proc {"Exception(#{exception}) with message matches to #{expected.inspect}"}) do
+            yield
+          end
+          m = ex.message
+        end
         msg = message(msg, "") {"Expected Exception(#{exception}) was raised, but the message doesn't match"}
 
         if assert == :assert_equal
-          assert_equal(expected, ex.message, msg)
+          assert_equal(expected, m, msg)
         else
-          msg = message(msg) { "Expected #{mu_pp expected} to match #{mu_pp ex.message}" }
-          assert expected =~ ex.message, msg
+          msg = message(msg) { "Expected #{mu_pp expected} to match #{mu_pp m}" }
+          assert expected =~ m, msg
           block.binding.eval("proc{|_|$~=_}").call($~)
         end
         ex
@@ -355,15 +363,41 @@ EOT
       #
       #    assert_respond_to("hello", :reverse)  #Succeeds
       #    assert_respond_to("hello", :does_not_exist)  #Fails
-      def assert_respond_to obj, (meth, priv), msg = nil
-        if priv
+      def assert_respond_to(obj, (meth, *priv), msg = nil)
+        unless priv.empty?
           msg = message(msg) {
-            "Expected #{mu_pp(obj)} (#{obj.class}) to respond to ##{meth}#{" privately" if priv}"
+            "Expected #{mu_pp(obj)} (#{obj.class}) to respond to ##{meth}#{" privately" if priv[0]}"
           }
-          return assert obj.respond_to?(meth, priv), msg
+          return assert obj.respond_to?(meth, *priv), msg
         end
         #get rid of overcounting
-        super if !caller[0].rindex(MINI_DIR, 0) || !obj.respond_to?(meth)
+        if caller_locations(1, 1)[0].path.start_with?(MINI_DIR)
+          return if obj.respond_to?(meth)
+        end
+        super(obj, meth, msg)
+      end
+
+      # :call-seq:
+      #   assert_not_respond_to( object, method, failure_message = nil )
+      #
+      #Tests if the given Object does not respond to +method+.
+      #
+      #An optional failure message may be provided as the final argument.
+      #
+      #    assert_not_respond_to("hello", :reverse)  #Fails
+      #    assert_not_respond_to("hello", :does_not_exist)  #Succeeds
+      def assert_not_respond_to(obj, (meth, *priv), msg = nil)
+        unless priv.empty?
+          msg = message(msg) {
+            "Expected #{mu_pp(obj)} (#{obj.class}) to not respond to ##{meth}#{" privately" if priv[0]}"
+          }
+          return assert !obj.respond_to?(meth, *priv), msg
+        end
+        #get rid of overcounting
+        if caller_locations(1, 1)[0].path.start_with?(MINI_DIR)
+          return unless obj.respond_to?(meth)
+        end
+        refute_respond_to(obj, meth, msg)
       end
 
       # :call-seq:
@@ -420,7 +454,7 @@ EOT
 
       ms = instance_methods(true).map {|sym| sym.to_s }
       ms.grep(/\Arefute_/) do |m|
-        mname = ('assert_not_' << m.to_s[/.*?_(.*)/, 1])
+        mname = ('assert_not_'.dup << m.to_s[/.*?_(.*)/, 1])
         alias_method(mname, m) unless ms.include? mname
       end
       alias assert_include assert_includes
@@ -446,52 +480,59 @@ EOT
         assert(failed.empty?, message(m) {failed.pretty_inspect})
       end
 
-      def assert_valid_syntax(code, fname = caller_locations(1, 1)[0], mesg = fname.to_s, verbose: nil)
-        code = code.b
-        code.sub!(/\A(?:\xef\xbb\xbf)?(\s*\#.*$)*(\n)?/n) {
-          "#$&#{"\n" if $1 && !$2}BEGIN{throw tag, :ok}\n"
-        }
-        code.force_encoding(Encoding::UTF_8)
+      # compatibility with test-unit
+      alias pend skip
+
+      if defined?(RubyVM::InstructionSequence)
+        def syntax_check(code, fname, line)
+          code = code.dup.force_encoding(Encoding::UTF_8)
+          RubyVM::InstructionSequence.compile(code, fname, fname, line)
+          :ok
+        end
+      else
+        def syntax_check(code, fname, line)
+          code = code.b
+          code.sub!(/\A(?:\xef\xbb\xbf)?(\s*\#.*$)*(\n)?/n) {
+            "#$&#{"\n" if $1 && !$2}BEGIN{throw tag, :ok}\n"
+          }
+          code = code.force_encoding(Encoding::UTF_8)
+          catch {|tag| eval(code, binding, fname, line - 1)}
+        end
+      end
+
+      def prepare_syntax_check(code, fname = caller_locations(2, 1)[0], mesg = fname.to_s, verbose: nil)
         verbose, $VERBOSE = $VERBOSE, verbose
-        yield if defined?(yield)
         case
         when Array === fname
           fname, line = *fname
         when defined?(fname.path) && defined?(fname.lineno)
           fname, line = fname.path, fname.lineno
         else
-          line = 0
-        end
-        assert_nothing_raised(SyntaxError, mesg) do
-          assert_equal(:ok, catch {|tag| eval(code, binding, fname, line)}, mesg)
+          line = 1
         end
+        yield(code, fname, line, mesg)
       ensure
         $VERBOSE = verbose
       end
 
-      def assert_syntax_error(code, error, fname = caller_locations(1, 1)[0], mesg = fname.to_s)
-        code = code.b
-        code.sub!(/\A(?:\xef\xbb\xbf)?(\s*\#.*$)*(\n)?/n) {
-          "#$&#{"\n" if $1 && !$2}BEGIN{throw tag, :ng}\n"
-        }
-        code.force_encoding(Encoding::US_ASCII)
-        verbose, $VERBOSE = $VERBOSE, nil
-        yield if defined?(yield)
-        case
-        when Array === fname
-          fname, line = *fname
-        when defined?(fname.path) && defined?(fname.lineno)
-          fname, line = fname.path, fname.lineno
-        else
-          line = 0
+      def assert_valid_syntax(code, *args)
+        prepare_syntax_check(code, *args) do |src, fname, line, mesg|
+          yield if defined?(yield)
+          assert_nothing_raised(SyntaxError, mesg) do
+            assert_equal(:ok, syntax_check(src, fname, line), mesg)
+          end
         end
-        e = assert_raise(SyntaxError, mesg) do
-          catch {|tag| eval(code, binding, fname, line)}
+      end
+
+      def assert_syntax_error(code, error, *args)
+        prepare_syntax_check(code, *args) do |src, fname, line, mesg|
+          yield if defined?(yield)
+          e = assert_raise(SyntaxError, mesg) do
+            syntax_check(src, fname, line)
+          end
+          assert_match(error, e.message, mesg)
+          e
         end
-        assert_match(error, e.message, mesg)
-        e
-      ensure
-        $VERBOSE = verbose
       end
 
       def assert_normal_exit(testsrc, message = '', child_env: nil, **opt)
@@ -513,14 +554,15 @@ EOT
             signame = Signal.signame(signo)
             sigdesc = "signal #{signo}"
           end
-          log = EnvUtil.diagnostic_reports(signame, EnvUtil.rubybin, pid, now)
+          log = EnvUtil.diagnostic_reports(signame, pid, now)
           if signame
             sigdesc = "SIG#{signame} (#{sigdesc})"
           end
           if status.coredump?
-            sigdesc << " (core dumped)"
+            sigdesc = "#{sigdesc} (core dumped)"
           end
-          full_message = ''
+          full_message = ''.dup
+          message = message.call if Proc === message
           if message and !message.empty?
             full_message << message << "\n"
           end
@@ -528,21 +570,22 @@ EOT
           full_message << " exit #{status.exitstatus}" if status.exited?
           full_message << " killed by #{sigdesc}" if sigdesc
           if out and !out.empty?
-            full_message << "\n#{out.b.gsub(/^/, '| ')}"
-            full_message << "\n" if /\n\z/ !~ full_message
+            full_message << "\n" << out.b.gsub(/^/, '| ')
+            full_message.sub!(/(?<!\n)\z/, "\n")
           end
           if log
-            full_message << "\n#{log.b.gsub(/^/, '| ')}"
+            full_message << "Diagnostic reports:\n" << log.b.gsub(/^/, '| ')
           end
           full_message
         end
         faildesc
       end
 
-      def assert_in_out_err(args, test_stdin = "", test_stdout = [], test_stderr = [], message = nil, **opt)
+      def assert_in_out_err(args, test_stdin = "", test_stdout = [], test_stderr = [], message = nil,
+                            success: nil, **opt)
         stdout, stderr, status = EnvUtil.invoke_ruby(args, test_stdin, true, true, **opt)
         if signo = status.termsig
-          EnvUtil.diagnostic_reports(Signal.signame(signo), EnvUtil.rubybin, status.pid, Time.now)
+          EnvUtil.diagnostic_reports(Signal.signame(signo), status.pid, Time.now)
         end
         if block_given?
           raise "test_stdout ignored, use block only or without block" if test_stdout != []
@@ -561,6 +604,15 @@ EOT
                 end
               end
             end
+            unless success.nil?
+              a.for("success?") do
+                if success
+                  assert_predicate(status, :success?)
+                else
+                  assert_not_predicate(status, :success?)
+                end
+              end
+            end
           end
           status
         end
@@ -582,21 +634,20 @@ EOT
           file ||= loc.path
           line ||= loc.lineno
         end
-        line -= 5 # lines until src
         src = <<eom
-# -*- coding: #{src.encoding}; -*-
+# -*- coding: #{line += __LINE__; src.encoding}; -*-
   require #{__dir__.dump};include Test::Unit::Assertions
   END {
     puts [Marshal.dump($!)].pack('m'), "assertions=\#{self._assertions}"
   }
-#{src}
+#{line -= __LINE__; src}
   class Test::Unit::Runner
     @@stop_auto_run = true
   end
 eom
         args = args.dup
         args.insert((Hash === args.first ? 1 : 0), "-w", "--disable=gems", *$:.map {|l| "-I#{l}"})
-        stdout, stderr, status = EnvUtil.invoke_ruby(args, src, true, true, timeout_error: nil, **opt)
+        stdout, stderr, status = EnvUtil.invoke_ruby(args, src, true, true, **opt)
         abort = status.coredump? || (status.signaled? && ABORT_SIGNALS.include?(status.termsig))
         assert(!abort, FailDesc[status, nil, stderr])
         self._assertions += stdout[/^assertions=(\d+)/, 1].to_i
@@ -614,7 +665,7 @@ eom
           else
             res.set_backtrace(caller)
           end
-          raise res
+          raise res unless SystemExit === res
         end
 
         # really is it succeed?
@@ -627,7 +678,11 @@ eom
       end
 
       def assert_warning(pat, msg = nil)
-        stderr = EnvUtil.verbose_warning { yield }
+        stderr = EnvUtil.with_default_internal(pat.encoding) {
+          EnvUtil.verbose_warning {
+            yield
+          }
+        }
         msg = message(msg) {diff pat, stderr}
         assert(pat === stderr, msg)
       end
@@ -637,14 +692,16 @@ eom
       end
 
       def assert_no_memory_leak(args, prepare, code, message=nil, limit: 2.0, rss: false, **opt)
-        require_relative 'memory_status'
+        require_relative '../../memory_status'
+        raise MiniTest::Skip, "unsupported platform" unless defined?(Memory::Status)
+
         token = "\e[7;1m#{$$.to_s}:#{Time.now.strftime('%s.%L')}:#{rand(0x10000).to_s(16)}:\e[m"
         token_dump = token.dump
         token_re = Regexp.quote(token)
         envs = args.shift if Array === args and Hash === args.first
         args = [
           "--disable=gems",
-          "-r", File.expand_path("../memory_status", __FILE__),
+          "-r", File.expand_path("../../../memory_status", __FILE__),
           *args,
           "-v", "-",
         ]
@@ -676,6 +733,27 @@ eom
         skip
       end
 
+      def assert_cpu_usage_low(msg = nil, pct: 0.01)
+        require 'benchmark'
+
+        tms = Benchmark.measure(msg || '') { yield }
+        max = pct * tms.real
+        if tms.real < 0.1 # TIME_QUANTUM_USEC in thread_pthread.c
+          warn "test #{msg || 'assert_cpu_usage_low'} too short to be accurate"
+        end
+
+        # kernel resolution can limit the minimum time we can measure
+        # [ruby-core:81540]
+        min_hz = windows? ? 67 : 100
+        min_measurable = 1.0 / min_hz
+        min_measurable *= 1.10 # add a little (10%) to account for misc. overheads
+        if max < min_measurable
+          max = min_measurable
+        end
+
+        assert_operator tms.total, :<=, max, msg
+      end
+
       def assert_is_minus_zero(f)
         assert(1.0/f == -Float::INFINITY, "#{f} is not -0.0")
       end
@@ -705,13 +783,13 @@ eom
               msg = message(msg) {
                 expect_msg = "Expected #{mu_pp pattern}\n"
                 if /\n[^\n]/ =~ rest
-                  actual_mesg = "to match\n"
+                  actual_mesg = +"to match\n"
                   rest.scan(/.*\n+/) {
                     actual_mesg << '  ' << $&.inspect << "+\n"
                   }
                   actual_mesg.sub!(/\+\n\z/, '')
                 else
-                  actual_mesg = "to match #{mu_pp rest}"
+                  actual_mesg = "to match " + mu_pp(rest)
                 end
                 actual_mesg << "\nafter #{i} patterns with #{actual.length - rest.length} characters"
                 expect_msg + actual_mesg
@@ -768,7 +846,7 @@ eom
           end
           result = File.__send__(predicate, *args)
           result = !result if neg
-          mesg = "Expected file " << args.shift.inspect
+          mesg = "Expected file ".dup << args.shift.inspect
           mesg << "#{neg} to be #{predicate}"
           mesg << mu_pp(args).sub(/\A\[(.*)\]\z/m, '(\1)') unless args.empty?
           mesg << " #{failure_message}" if failure_message
@@ -796,12 +874,24 @@ eom
           @failures[key] = [@count, e]
         end
 
+        def foreach(*keys)
+          keys.each do |key|
+            @count += 1
+            begin
+              yield key
+            rescue Exception => e
+              @failures[key] = [@count, e]
+            end
+          end
+        end
+
         def message
           i = 0
           total = @count.to_s
           fmt = "%#{total.size}d"
           @failures.map {|k, (n, v)|
-            "\n#{i+=1}. [#{fmt%n}/#{total}] Assertion for #{k.inspect}\n#{v.message.gsub(/^/, '   | ')}"
+            v = v.message
+            "\n#{i+=1}. [#{fmt%n}/#{total}] Assertion for #{k.inspect}\n#{v.b.gsub(/^/, '   | ').force_encoding(v.encoding)}"
           }.join("\n")
         end
 
@@ -810,12 +900,21 @@ eom
         end
       end
 
-      def all_assertions(msg = nil)
+      def assert_all_assertions(msg = nil)
         all = AllFailures.new
         yield all
       ensure
-        assert(all.pass?, message(msg) {all.message})
+        assert(all.pass?, message(msg) {all.message.chomp(".")})
       end
+      alias all_assertions assert_all_assertions
+
+      def assert_all_assertions_foreach(msg = nil, *keys, &block)
+        all = AllFailures.new
+        all.foreach(*keys, &block)
+      ensure
+        assert(all.pass?, message(msg) {all.message.chomp(".")})
+      end
+      alias all_assertions_foreach assert_all_assertions_foreach
 
       def build_message(head, template=nil, *arguments) #:nodoc:
         template &&= template.chomp
