@@ -111,6 +111,12 @@ class HTTPHeaderTest < Test::Unit::TestCase
     assert_equal ['test string'], @c.get_fields('my-header')
   end
 
+  class D; include Net::HTTPHeader; end
+
+  def test_nil_variable_header
+    assert_nothing_raised { D.new.initialize_http_header({Authorization: nil}) }
+  end
+
   def test_delete
     @c['My-Header'] = 'test'
     assert_equal 'test', @c['My-Header']
@@ -134,6 +140,12 @@ class HTTPHeaderTest < Test::Unit::TestCase
       assert_equal 'my-header', k
       assert_equal 'test', v
     end
+    e = @c.each
+    assert_equal 1, e.size
+    e.each do |k, v|
+      assert_equal 'my-header', k
+      assert_equal 'test', v
+    end
   end
 
   def test_each_key
@@ -144,6 +156,26 @@ class HTTPHeaderTest < Test::Unit::TestCase
     @c.each_key do |k|
       assert_equal 'my-header', k
     end
+    e = @c.each_key
+    assert_equal 1, e.size
+    e.each do |k|
+      assert_equal 'my-header', k
+    end
+  end
+
+  def test_each_capitalized_name
+    @c['my-header'] = 'test'
+    @c.each_capitalized_name do |k|
+      assert_equal 'My-Header', k
+    end
+    @c.each_capitalized_name do |k|
+      assert_equal 'My-Header', k
+    end
+    e = @c.each_capitalized_name
+    assert_equal 1, e.size
+    e.each do |k|
+      assert_equal 'My-Header', k
+    end
   end
 
   def test_each_value
@@ -154,6 +186,11 @@ class HTTPHeaderTest < Test::Unit::TestCase
     @c.each_value do |v|
       assert_equal 'test', v
     end
+    e = @c.each_value
+    assert_equal 1, e.size
+    e.each do |v|
+      assert_equal 'test', v
+    end
   end
 
   def test_canonical_each
@@ -162,6 +199,12 @@ class HTTPHeaderTest < Test::Unit::TestCase
       assert_equal 'My-Header', k
       assert_equal 'a, b', v
     end
+    e = @c.canonical_each
+    assert_equal 1, e.size
+    e.each do |k,v|
+      assert_equal 'My-Header', k
+      assert_equal 'a, b', v
+    end
   end
 
   def test_each_capitalized
@@ -170,6 +213,12 @@ class HTTPHeaderTest < Test::Unit::TestCase
       assert_equal 'My-Header', k
       assert_equal 'a, b', v
     end
+    e = @c.each_capitalized
+    assert_equal 1, e.size
+    e.each do |k,v|
+      assert_equal 'My-Header', k
+      assert_equal 'a, b', v
+    end
   end
 
   def test_each_capitalized_with_symbol
@@ -178,6 +227,12 @@ class HTTPHeaderTest < Test::Unit::TestCase
       assert_equal "My_header", k
       assert_equal 'a, b', v
     end
+    e = @c.each_capitalized
+    assert_equal 1, e.size
+    e.each do |k,v|
+      assert_equal 'My_header', k
+      assert_equal 'a, b', v
+    end
   end
 
   def test_key?
