@@ -1,11 +1,17 @@
+# frozen_string_literal: true
 require 'net/smtp'
-require 'minitest/autorun'
+require 'test/unit'
 
 module Net
-  class TestSSLSocket < MiniTest::Unit::TestCase
+  class TestSSLSocket < Test::Unit::TestCase
     class MySMTP < SMTP
       attr_accessor :fake_tcp, :fake_ssl
 
+      def initialize(*args)
+        super(*args)
+        @open_timeout = nil
+      end
+
       def tcp_socket address, port
         fake_tcp
       end
@@ -15,6 +21,7 @@ module Net
       end
     end
 
+    require 'stringio'
     class SSLSocket < StringIO
       attr_accessor :sync_close, :connected, :closed
 
@@ -51,7 +58,7 @@ module Net
       connection.fake_tcp = tcp_socket
       connection.fake_ssl = ssl_socket
 
-      assert_raises(OpenSSL::SSL::SSLError) do
+      assert_raise(OpenSSL::SSL::SSLError) do
         connection.start
       end
       assert_equal true, ssl_socket.closed
@@ -87,4 +94,4 @@ module Net
       ].join("\r\n") + "\r\n"
     end
   end
-end
+end if defined?(OpenSSL)
