@@ -1,6 +1,6 @@
+# frozen_string_literal: true
 begin
   require_relative 'dummyparser'
-  require_relative '../ruby/envutil'
   require 'test/unit'
   ripper_test = true
   module TestRipper; end
@@ -9,13 +9,10 @@ end
 
 class TestRipper::ParserEvents < Test::Unit::TestCase
 
-  # should be enabled
   def test_event_coverage
-    dispatched = Ripper::PARSER_EVENTS.map {|event,*| event }
-    dispatched.each do |e|
-      assert_equal true, respond_to?("test_#{e}", true),
-                   "event not tested: #{e.inspect}"
-    end
+    dispatched = Ripper::PARSER_EVENTS
+    tested = self.class.instance_methods(false).grep(/\Atest_(\w+)/) {$1.intern}
+    assert_empty dispatched-tested
   end
 
   def parse(str, nm = nil, &bl)
@@ -24,6 +21,18 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     dp.parse.to_s
   end
 
+  def compile_error(str)
+    parse(str, :compile_error) {|e, msg| return msg}
+  end
+
+  def warning(str)
+    parse(str, :warning) {|e, *args| return args}
+  end
+
+  def warn(str)
+    parse(str, :warn) {|e, *args| return args}
+  end
+
   def test_program
     thru_program = false
     assert_equal '[void()]', parse('', :on_program) {thru_program = true}
@@ -46,11 +55,15 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
   end
 
   def test_var_ref
-    assert_equal '[ref(a)]', parse('a')
+    assert_equal '[assign(var_field(a),ref(a))]', parse('a=a')
     assert_equal '[ref(nil)]', parse('nil')
     assert_equal '[ref(true)]', parse('true')
   end
 
+  def test_vcall
+    assert_equal '[vcall(a)]', parse('a')
+  end
+
   def test_BEGIN
     assert_equal '[BEGIN([void()])]', parse('BEGIN{}')
     assert_equal '[BEGIN([ref(nil)])]', parse('BEGIN{nil}')
@@ -77,15 +90,15 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     assert_equal '[fcall(m,[])]', parse('m()')
     assert_equal '[fcall(m,[1])]', parse('m(1)')
     assert_equal '[fcall(m,[1,2])]', parse('m(1,2)')
-    assert_equal '[fcall(m,[*ref(r)])]', parse('m(*r)')
-    assert_equal '[fcall(m,[1,*ref(r)])]', parse('m(1,*r)')
-    assert_equal '[fcall(m,[1,2,*ref(r)])]', parse('m(1,2,*r)')
-    assert_equal '[fcall(m,[&ref(r)])]', parse('m(&r)')
-    assert_equal '[fcall(m,[1,&ref(r)])]', parse('m(1,&r)')
-    assert_equal '[fcall(m,[1,2,&ref(r)])]', parse('m(1,2,&r)')
-    assert_equal '[fcall(m,[*ref(a),&ref(b)])]', parse('m(*a,&b)')
-    assert_equal '[fcall(m,[1,*ref(a),&ref(b)])]', parse('m(1,*a,&b)')
-    assert_equal '[fcall(m,[1,2,*ref(a),&ref(b)])]', parse('m(1,2,*a,&b)')
+    assert_equal '[fcall(m,[*vcall(r)])]', parse('m(*r)')
+    assert_equal '[fcall(m,[1,*vcall(r)])]', parse('m(1,*r)')
+    assert_equal '[fcall(m,[1,2,*vcall(r)])]', parse('m(1,2,*r)')
+    assert_equal '[fcall(m,[&vcall(r)])]', parse('m(&r)')
+    assert_equal '[fcall(m,[1,&vcall(r)])]', parse('m(1,&r)')
+    assert_equal '[fcall(m,[1,2,&vcall(r)])]', parse('m(1,2,&r)')
+    assert_equal '[fcall(m,[*vcall(a),&vcall(b)])]', parse('m(*a,&b)')
+    assert_equal '[fcall(m,[1,*vcall(a),&vcall(b)])]', parse('m(1,*a,&b)')
+    assert_equal '[fcall(m,[1,2,*vcall(a),&vcall(b)])]', parse('m(1,2,*a,&b)')
   end
 
   def test_args_add
@@ -120,8 +133,8 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
   end
 
   def test_aref
-    assert_equal '[aref(ref(v),[1])]', parse('v[1]')
-    assert_equal '[aref(ref(v),[1,2])]', parse('v[1,2]')
+    assert_equal '[aref(vcall(v),[1])]', parse('v[1]')
+    assert_equal '[aref(vcall(v),[1,2])]', parse('v[1,2]')
   end
 
   def test_assoclist_from_args
@@ -142,8 +155,14 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     assert_equal true, thru_assoc_new
   end
 
+  def test_assoc_splat
+    thru_assoc_splat = false
+    parse('m(**h)', :on_assoc_splat) {thru_assoc_splat = true}
+    assert_equal true, thru_assoc_splat
+  end
+
   def test_aref_field
-    assert_equal '[assign(aref_field(ref(a),[1]),2)]', parse('a[1]=2')
+    assert_equal '[assign(aref_field(vcall(a),[1]),2)]', parse('a[1]=2')
   end
 
   def test_arg_ambiguous
@@ -154,12 +173,20 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
 
   def test_operator_ambiguous
     thru_operator_ambiguous = false
-    parse('a=1; a %[]', :on_operator_ambiguous) {thru_operator_ambiguous = true}
+    token = syntax = nil
+    parse('a=1; a %[]', :on_operator_ambiguous) {|*a|
+      thru_operator_ambiguous = true
+      _, token, syntax = *a
+    }
     assert_equal true, thru_operator_ambiguous
+    assert_equal :%, token
+    assert_equal "string literal", syntax
   end
 
   def test_array   # array literal
     assert_equal '[array([1,2,3])]', parse('[1,2,3]')
+    assert_equal '[array([abc,def])]', parse('%w[abc def]')
+    assert_equal '[array([abc,def])]', parse('%W[abc def]')
   end
 
   def test_assign   # generic assignment
@@ -168,28 +195,73 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
 
   def test_assign_error
     thru_assign_error = false
-    parse('$` = 1', :on_assign_error) {thru_assign_error = true}
+    result = parse('self = 1', :on_assign_error) {thru_assign_error = true}
     assert_equal true, thru_assign_error
+    assert_equal '[assign(assign_error(var_field(self)),1)]', result
+  end
+
+  def test_assign_error_backref
     thru_assign_error = false
-    parse('$`, _ = 1', :on_assign_error) {thru_assign_error = true}
+    result =
+      parse('$` = 1', :on_assign_error) {thru_assign_error = true}
     assert_equal true, thru_assign_error
+    assert_equal '[assign(assign_error(var_field($`)),1)]', result
 
     thru_assign_error = false
-    parse('self::X = 1', :on_assign_error) {thru_assign_error = true}
+    result =
+      parse('$`, _ = 1', :on_assign_error) {thru_assign_error = true}
+    assert_equal true, thru_assign_error
+    assert_equal '[massign([assign_error(var_field($`)),var_field(_)],1)]', result
+  end
+
+  def test_assign_error_const_qualified
+    thru_assign_error = false
+    result =
+      parse('self::X = 1', :on_assign_error) {thru_assign_error = true}
     assert_equal false, thru_assign_error
-    parse('def m\n self::X = 1\nend', :on_assign_error) {thru_assign_error = true}
-    assert_equal true, thru_assign_error
+    assert_equal "[assign(const_path_field(ref(self),X),1)]", result
 
     thru_assign_error = false
-    parse('X = 1', :on_assign_error) {thru_assign_error = true}
-    assert_equal false, thru_assign_error
-    parse('def m\n X = 1\nend', :on_assign_error) {thru_assign_error = true}
+    result =
+      parse("def m\n self::X = 1\nend", :on_assign_error) {thru_assign_error = true}
     assert_equal true, thru_assign_error
+    assert_include result, "assign_error(const_path_field(ref(self),X))"
 
+    thru_assign_error = false
+    result =
+      parse("def m\n self::X, a = 1, 2\nend", :on_assign_error) {thru_assign_error = true}
+    assert_equal true, thru_assign_error
+    assert_include result, "assign_error(const_path_field(ref(self),X))"
+  end
+
+  def test_assign_error_const
+    thru_assign_error = false
+    result = parse('X = 1', :on_assign_error) {thru_assign_error = true}
+    assert_equal false, thru_assign_error
+    assert_equal "[assign(var_field(X),1)]", result
+
+    thru_assign_error = false
+    result = parse('X, a = 1, 2', :on_assign_error) {thru_assign_error = true}
+    assert_equal false, thru_assign_error
+    assert_include result, "massign([var_field(X),var_field(a)],"
+
+    result = parse("def m\n X = 1\nend", :on_assign_error) {thru_assign_error = true}
+    assert_equal true, thru_assign_error
+    assert_include result, "assign_error(var_field(X))"
+    thru_assign_error = false
+    result = parse("def m\n X, a = 1, 2\nend", :on_assign_error) {thru_assign_error = true}
+    assert_equal true, thru_assign_error
+    assert_include result, "assign_error(var_field(X))"
+  end
+
+  def test_assign_error_const_toplevel
     thru_assign_error = false
     parse('::X = 1', :on_assign_error) {thru_assign_error = true}
     assert_equal false, thru_assign_error
-    parse('def m\n ::X = 1\nend', :on_assign_error) {thru_assign_error = true}
+    parse("def m\n ::X = 1\nend", :on_assign_error) {thru_assign_error = true}
+    assert_equal true, thru_assign_error
+    thru_assign_error = false
+    parse("def m\n ::X, a = 1, 2\nend", :on_assign_error) {thru_assign_error = true}
     assert_equal true, thru_assign_error
   end
 
@@ -214,9 +286,8 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     assert_equal true, thru_begin
   end
 
-  def test_binary
-    thru_binary = nil
-    %w"and or + - * / % ** | ^ & <=> > >= < <= == === != =~ !~ << >> && ||".each do |op|
+  %w"and or + - * / % ** | ^ & <=> > >= < <= == === != =~ !~ << >> && ||".each do |op|
+    define_method("test_binary(#{op})") do
       thru_binary = false
       parse("a #{op} b", :on_binary) {thru_binary = true}
       assert_equal true, thru_binary
@@ -316,12 +387,51 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     }
     assert_equal true, thru_call
     assert_equal "[call(ref(self),.,foo)]", tree
+
+    thru_call = false
+    assert_nothing_raised {
+      tree = parse("self.foo()", :on_call) {thru_call = true}
+    }
+    assert_equal true, thru_call
+    assert_equal "[call(ref(self),.,foo,[])]", tree
+
     thru_call = false
     assert_nothing_raised(bug2233) {
       tree = parse("foo.()", :on_call) {thru_call = true}
     }
     assert_equal true, thru_call
-    assert_equal "[call(ref(foo),.,call,[])]", tree
+    assert_equal "[call(vcall(foo),.,call,[])]", tree
+
+    thru_call = false
+    assert_nothing_raised {
+      tree = parse("self::foo", :on_call) {thru_call = true}
+    }
+    assert_equal true, thru_call
+    assert_equal "[call(ref(self),::,foo)]", tree
+
+    thru_call = false
+    assert_nothing_raised {
+      tree = parse("self::foo()", :on_call) {thru_call = true}
+    }
+    assert_equal true, thru_call
+    assert_equal "[call(ref(self),::,foo,[])]", tree
+
+    thru_call = false
+    assert_nothing_raised(bug2233) {
+      tree = parse("foo::()", :on_call) {thru_call = true}
+    }
+    assert_equal true, thru_call
+    assert_equal "[call(vcall(foo),::,call,[])]", tree
+
+    thru_call = false
+    tree = parse("self&.foo", :on_call) {thru_call = true}
+    assert_equal true, thru_call
+    assert_equal "[call(ref(self),&.,foo)]", tree
+
+    thru_call = false
+    tree = parse("self&.foo()", :on_call) {thru_call = true}
+    assert_equal true, thru_call
+    assert_equal "[call(ref(self),&.,foo,[])]", tree
   end
 
   def test_excessed_comma
@@ -343,17 +453,30 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
   def test_heredoc
     bug1921 = '[ruby-core:24855]'
     thru_heredoc_beg = false
-    tree = parse("<<EOS\nheredoc\nEOS\n", :on_heredoc_beg) {thru_heredoc_beg = true}
+    tree = parse("<""<EOS\nheredoc\nEOS\n", :on_heredoc_beg) {thru_heredoc_beg = true}
     assert_equal true, thru_heredoc_beg
     assert_match(/string_content\(\),heredoc\n/, tree, bug1921)
     heredoc = nil
-    parse("<<EOS\nheredoc1\nheredoc2\nEOS\n", :on_string_add) {|e, n, s| heredoc = s}
+    parse("<""<EOS\nheredoc1\nheredoc2\nEOS\n", :on_string_add) {|e, n, s| heredoc = s}
     assert_equal("heredoc1\nheredoc2\n", heredoc, bug1921)
     heredoc = nil
-    parse("<<-EOS\nheredoc1\nheredoc2\n\tEOS\n", :on_string_add) {|e, n, s| heredoc = s}
+    parse("<""<-EOS\nheredoc1\nheredoc2\n\tEOS\n", :on_string_add) {|e, n, s| heredoc = s}
     assert_equal("heredoc1\nheredoc2\n", heredoc, bug1921)
   end
 
+  def test_heredoc_dedent
+    thru_heredoc_dedent = false
+    str = width = nil
+    tree = parse("<""<~EOS\n heredoc\nEOS\n", :on_heredoc_dedent) {|e, s, w|
+      thru_heredoc_dedent = true
+      str = s
+      width = w
+    }
+    assert_equal true, thru_heredoc_dedent
+    assert_match(/string_content\(\), heredoc\n/, tree)
+    assert_equal(1, width)
+  end
+
   def test_massign
     thru_massign = false
     parse("a, b = 1, 2", :on_massign) {thru_massign = true}
@@ -368,15 +491,51 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
 
   def test_mlhs_add_star
     bug2232 = '[ruby-core:26163]'
+    bug4364 = '[ruby-core:35078]'
 
     thru_mlhs_add_star = false
     tree = parse("a, *b = 1, 2", :on_mlhs_add_star) {thru_mlhs_add_star = true}
     assert_equal true, thru_mlhs_add_star
-    assert_match(/mlhs_add_star\(mlhs_add\(mlhs_new\(\),a\),b\)/, tree)
+    assert_include(tree, "massign([var_field(a),*var_field(b)]")
     thru_mlhs_add_star = false
     tree = parse("a, *b, c = 1, 2", :on_mlhs_add_star) {thru_mlhs_add_star = true}
     assert_equal true, thru_mlhs_add_star
-    assert_match(/mlhs_add\(mlhs_add_star\(mlhs_add\(mlhs_new\(\),a\),b\),mlhs_add\(mlhs_new\(\),c\)\)/, tree, bug2232)
+    assert_include(tree, "massign([var_field(a),*var_field(b),var_field(c)]", bug2232)
+    thru_mlhs_add_star = false
+    tree = parse("a, *, c = 1, 2", :on_mlhs_add_star) {thru_mlhs_add_star = true}
+    assert_equal true, thru_mlhs_add_star
+    assert_include(tree, "massign([var_field(a),*,var_field(c)]", bug4364)
+    thru_mlhs_add_star = false
+    tree = parse("*b, c = 1, 2", :on_mlhs_add_star) {thru_mlhs_add_star = true}
+    assert_equal true, thru_mlhs_add_star
+    assert_include(tree, "massign([*var_field(b),var_field(c)]", bug4364)
+    thru_mlhs_add_star = false
+    tree = parse("*, c = 1, 2", :on_mlhs_add_star) {thru_mlhs_add_star = true}
+    assert_equal true, thru_mlhs_add_star
+    assert_include(tree, "massign([*,var_field(c)],", bug4364)
+  end
+
+  def test_mlhs_add_post
+    thru_mlhs_add_post = false
+    tree = parse("a, *b = 1, 2", :on_mlhs_add_post) {thru_mlhs_add_post = true}
+    assert_equal false, thru_mlhs_add_post
+    assert_include(tree, "massign([var_field(a),*var_field(b)],")
+    thru_massign_add_post = false
+    tree = parse("a, *b, c = 1, 2", :on_mlhs_add_post) {thru_mlhs_add_post = true}
+    assert_equal true, thru_mlhs_add_post
+    assert_include(tree, "massign([var_field(a),*var_field(b),var_field(c)],")
+    thru_mlhs_add_post = false
+    tree = parse("a, *, c = 1, 2", :on_mlhs_add_post) {thru_mlhs_add_post = true}
+    assert_equal true, thru_mlhs_add_post
+    assert_include(tree, "massign([var_field(a),*,var_field(c)],")
+    thru_mlhs_add_post = false
+    tree = parse("*b, c = 1, 2", :on_mlhs_add_post) {thru_mlhs_add_post = true}
+    assert_equal true, thru_mlhs_add_post
+    assert_include(tree, "massign([*var_field(b),var_field(c)],")
+    thru_mlhs_add_post = false
+    tree = parse("*, c = 1, 2", :on_mlhs_add_post) {thru_mlhs_add_post = true}
+    assert_equal true, thru_mlhs_add_post
+    assert_include(tree, "massign([*,var_field(c)],")
   end
 
   def test_mlhs_new
@@ -474,8 +633,13 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
 
   def test_defs
     thru_defs = false
-    parse('def foo.bar; end', :on_defs) {thru_defs = true}
+    tree = parse('def foo.bar; end', :on_defs) {thru_defs = true}
     assert_equal true, thru_defs
+    assert_equal("[defs(vcall(foo),.,bar,[],bodystmt([void()]))]", tree)
+
+    thru_parse_error = false
+    tree = parse('def foo&.bar; end', :on_parse_error) {thru_parse_error = true}
+    assert_equal(true, thru_parse_error)
   end
 
   def test_do_block
@@ -500,6 +664,10 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     thru_dyna_symbol = false
     parse(':"#{foo}"', :on_dyna_symbol) {thru_dyna_symbol = true}
     assert_equal true, thru_dyna_symbol
+
+    thru_dyna_symbol = false
+    parse('{"#{foo}": 1}', :on_dyna_symbol) {thru_dyna_symbol = true}
+    assert_equal true, thru_dyna_symbol
   end
 
   def test_else
@@ -573,8 +741,8 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
 
   def test_magic_comment
     thru_magic_comment = false
-    parse('# -*- foo:bar -*-', :on_magic_comment) {thru_magic_comment = true}
-    assert_equal true, thru_magic_comment
+    parse('# -*- bug-5753: ruby-dev:44984 -*-', :on_magic_comment) {|*x|thru_magic_comment = x}
+    assert_equal [:on_magic_comment, "bug_5753", "ruby-dev:44984"], thru_magic_comment
   end
 
   def test_method_add_block
@@ -639,47 +807,73 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
 
   def test_opassign
     thru_opassign = false
-    parse('a += b', :on_opassign) {thru_opassign = true}
+    tree = parse('a += b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),+=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a -= b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),-=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a *= b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),*=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a /= b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),/=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a %= b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),%=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a **= b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),**=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a &= b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),&=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a |= b', :on_opassign) {thru_opassign = true}
+    assert_equal "[opassign(var_field(a),|=,vcall(b))]", tree
     assert_equal true, thru_opassign
     thru_opassign = false
-    parse('a -= b', :on_opassign) {thru_opassign = true}
+    tree = parse('a <<= b', :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),<<=,vcall(b))]", tree
     thru_opassign = false
-    parse('a *= b', :on_opassign) {thru_opassign = true}
+    tree = parse('a >>= b', :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),>>=,vcall(b))]", tree
     thru_opassign = false
-    parse('a /= b', :on_opassign) {thru_opassign = true}
+    tree = parse('a &&= b', :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),&&=,vcall(b))]", tree
     thru_opassign = false
-    parse('a %= b', :on_opassign) {thru_opassign = true}
+    tree = parse('a ||= b', :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),||=,vcall(b))]", tree
     thru_opassign = false
-    parse('a **= b', :on_opassign) {thru_opassign = true}
+    tree = parse('a::X ||= c 1', :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(const_path_field(vcall(a),X),||=,command(c,[1]))]", tree
+
     thru_opassign = false
-    parse('a &= b', :on_opassign) {thru_opassign = true}
+    tree = parse("self.foo += 1", :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(field(ref(self),.,foo),+=,1)]", tree
+
     thru_opassign = false
-    parse('a |= b', :on_opassign) {thru_opassign = true}
-    assert_equal true, thru_opassign
-    thru_opassign = false
-    parse('a <<= b', :on_opassign) {thru_opassign = true}
-    assert_equal true, thru_opassign
-    thru_opassign = false
-    parse('a >>= b', :on_opassign) {thru_opassign = true}
-    assert_equal true, thru_opassign
-    thru_opassign = false
-    parse('a &&= b', :on_opassign) {thru_opassign = true}
-    assert_equal true, thru_opassign
-    thru_opassign = false
-    parse('a ||= b', :on_opassign) {thru_opassign = true}
+    tree = parse("self&.foo += 1", :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(field(ref(self),&.,foo),+=,1)]", tree
   end
 
   def test_opassign_error
     thru_opassign = []
-    events = [:on_opassign, :on_assign_error]
-    parse('a::X ||= c 1', events) {|a,*b|
+    events = [:on_opassign]
+    parse('$~ ||= 1', events) {|a,*b|
       thru_opassign << a
     }
     assert_equal events, thru_opassign
@@ -701,15 +895,109 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
   end
 
   def test_params
+    arg = nil
     thru_params = false
-    parse('a {||}', :on_params) {thru_params = true}
+    parse('a {||}', :on_params) {|_, *v| thru_params = true; arg = v}
     assert_equal true, thru_params
+    assert_equal [nil, nil, nil, nil, nil, nil, nil], arg
     thru_params = false
-    parse('a {|x|}', :on_params) {thru_params = true}
+    parse('a {|x|}', :on_params) {|_, *v| thru_params = true; arg = v}
     assert_equal true, thru_params
+    assert_equal [["x"], nil, nil, nil, nil, nil, nil], arg
     thru_params = false
-    parse('a {|*x|}', :on_params) {thru_params = true}
+    parse('a {|*x|}', :on_params) {|_, *v| thru_params = true; arg = v}
     assert_equal true, thru_params
+    assert_equal [nil, nil, "*x", nil, nil, nil, nil], arg
+    thru_params = false
+    parse('a {|x: 1|}', :on_params) {|_, *v| thru_params = true; arg = v}
+    assert_equal true, thru_params
+    assert_equal [nil, nil, nil, nil, [["x:", "1"]], nil, nil], arg
+    thru_params = false
+    parse('a {|x:|}', :on_params) {|_, *v| thru_params = true; arg = v}
+    assert_equal true, thru_params
+    assert_equal [nil, nil, nil, nil, [["x:", false]], nil, nil], arg
+    thru_params = false
+    parse('a {|**x|}', :on_params) {|_, *v| thru_params = true; arg = v}
+    assert_equal true, thru_params
+    assert_equal [nil, nil, nil, nil, nil, "**x", nil], arg
+  end
+
+  def test_params_mlhs
+    thru_mlhs = false
+    tree = parse("proc {|(a, b)|}", :on_mlhs_paren) {thru_mlhs = true}
+    assert_equal true, thru_mlhs
+    assert_include(tree, "[mlhs([a,b])]")
+  end
+
+  def test_params_mlhs_add
+    thru_mlhs_add = false
+    tree = parse("proc {|(a, b)|}", :on_mlhs_add) {thru_mlhs_add = true}
+    assert_equal true, thru_mlhs_add
+    assert_include(tree, "[mlhs([a,b])]")
+  end
+
+  def test_params_mlhs_add_star
+    thru_mlhs_add_star = false
+    tree = parse("proc {|(a, *b)|}", :on_mlhs_add_star) {thru_mlhs_add_star = true}
+    assert_equal true, thru_mlhs_add_star
+    assert_include(tree, "[mlhs([a,*b])]")
+    thru_mlhs_add_star = false
+    tree = parse("proc {|(a, *b, c)|}", :on_mlhs_add_star) {thru_mlhs_add_star = true}
+    assert_equal true, thru_mlhs_add_star
+    assert_include(tree, "[mlhs([a,*b,c])]")
+    thru_mlhs_add_star = false
+    tree = parse("proc {|(a, *, c)|}", :on_mlhs_add_star) {thru_mlhs_add_star = true}
+    assert_equal true, thru_mlhs_add_star
+    assert_include(tree, "[mlhs([a,*,c])]")
+    thru_mlhs_add_star = false
+    tree = parse("proc {|(*b, c)|}", :on_mlhs_add_star) {thru_mlhs_add_star = true}
+    assert_equal true, thru_mlhs_add_star
+    assert_include(tree, "[mlhs([*b,c])]")
+    thru_mlhs_add_star = false
+    tree = parse("proc {|(*b)|}", :on_mlhs_add_star) {thru_mlhs_add_star = true}
+    assert_equal true, thru_mlhs_add_star
+    assert_include(tree, "[mlhs([*b])]")
+  end
+
+  def test_params_mlhs_add_post
+    thru_mlhs_add_post = false
+    tree = parse("proc {|(a, *b)|}", :on_mlhs_add_post) {thru_mlhs_add_post = true}
+    assert_equal false, thru_mlhs_add_post
+    assert_include(tree, "mlhs([a,*b])")
+    thru_mlhs_add_post = false
+    tree = parse("proc {|(a, *b, c)|}", :on_mlhs_add_post) {thru_mlhs_add_post = true}
+    assert_equal true, thru_mlhs_add_post
+    assert_include(tree, "mlhs([a,*b,c])")
+    thru_mlhs_add_post = false
+    tree = parse("proc {|(a, *, c)|}", :on_mlhs_add_post) {thru_mlhs_add_post = true}
+    assert_equal true, thru_mlhs_add_post
+    assert_include(tree, "mlhs([a,*,c])")
+    thru_mlhs_add_post = false
+    tree = parse("proc {|(*b, c)|}", :on_mlhs_add_post) {thru_mlhs_add_post = true}
+    assert_equal true, thru_mlhs_add_post
+    assert_include(tree, "mlhs([*b,c])")
+    thru_mlhs_add_post = false
+    tree = parse("proc {|(*, c)|}", :on_mlhs_add_post) {thru_mlhs_add_post = true}
+    assert_equal true, thru_mlhs_add_post
+    assert_include(tree, "mlhs([*,c])")
+  end
+
+  def test_params_mlhs_new
+    thru_mlhs_new = false
+    tree = parse("proc {|(a, b)|}", :on_mlhs_new) {thru_mlhs_new = true}
+    assert_equal true, thru_mlhs_new
+    assert_include(tree, "[mlhs([a,b])]")
+  end
+
+  def test_params_mlhs_paren
+    thru_mlhs_paren = 0
+    tree = parse("proc {|(a, b)|}", :on_mlhs_paren) {thru_mlhs_paren += 1}
+    assert_equal 1, thru_mlhs_paren
+    assert_include(tree, "[mlhs([a,b])]")
+    thru_mlhs_paren = 0
+    tree = parse("proc {|((a, b))|}", :on_mlhs_paren) {thru_mlhs_paren += 1}
+    assert_equal 2, thru_mlhs_paren
+    assert_include(tree, "[mlhs([a,b])]")
   end
 
   def test_paren
@@ -726,8 +1014,35 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
 
   def test_qwords_add
     thru_qwords_add = false
-    parse('%w[a]', :on_qwords_add) {thru_qwords_add = true}
+    tree = parse('%w[a]', :on_qwords_add) {thru_qwords_add = true}
     assert_equal true, thru_qwords_add
+    assert_equal '[array([a])]', tree
+    thru_qwords_add = false
+    tree = parse('%w[ a ]', :on_qwords_add) {thru_qwords_add = true}
+    assert_equal true, thru_qwords_add
+    assert_equal '[array([a])]', tree
+  end
+
+  def test_qsymbols_add
+    thru_qsymbols_add = false
+    tree = parse('%i[a]', :on_qsymbols_add) {thru_qsymbols_add = true}
+    assert_equal true, thru_qsymbols_add
+    assert_equal '[array([:a])]', tree
+    thru_qsymbols_add = false
+    tree = parse('%i[ a ]', :on_qsymbols_add) {thru_qsymbols_add = true}
+    assert_equal true, thru_qsymbols_add
+    assert_equal '[array([:a])]', tree
+  end
+
+  def test_symbols_add
+    thru_symbols_add = false
+    tree = parse('%I[a]', :on_symbols_add) {thru_symbols_add = true}
+    assert_equal true, thru_symbols_add
+    assert_equal '[array([:a])]', tree
+    thru_symbols_add = false
+    tree = parse('%I[ a ]', :on_symbols_add) {thru_symbols_add = true}
+    assert_equal true, thru_symbols_add
+    assert_equal '[array([:a])]', tree
   end
 
   def test_qwords_new
@@ -736,6 +1051,18 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     assert_equal true, thru_qwords_new
   end
 
+  def test_qsymbols_new
+    thru_qsymbols_new = false
+    parse('%i[]', :on_qsymbols_new) {thru_qsymbols_new = true}
+    assert_equal true, thru_qsymbols_new
+  end
+
+  def test_symbols_new
+    thru_symbols_new = false
+    parse('%I[]', :on_symbols_new) {thru_symbols_new = true}
+    assert_equal true, thru_symbols_new
+  end
+
   def test_redo
     thru_redo = false
     parse('redo', :on_redo) {thru_redo = true}
@@ -762,14 +1089,26 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
 
   def test_rescue
     thru_rescue = false
-    parse('begin; rescue; end', :on_rescue) {thru_rescue = true}
+    parsed = parse('begin; 1; rescue => e; 2; end', :on_rescue) {thru_rescue = true}
     assert_equal true, thru_rescue
+    assert_match(/1.*rescue/, parsed)
+    assert_match(/rescue\(,var_field\(e\),\[2\]\)/, parsed)
+  end
+
+  def test_rescue_class
+    thru_rescue = false
+    parsed = parse('begin; 1; rescue RuntimeError => e; 2; end', :on_rescue) {thru_rescue = true}
+    assert_equal true, thru_rescue
+    assert_match(/1.*rescue/, parsed)
+    assert_match(/rescue\(\[ref\(RuntimeError\)\],var_field\(e\),\[2\]\)/, parsed)
   end
 
   def test_rescue_mod
     thru_rescue_mod = false
-    parse('nil rescue nil', :on_rescue_mod) {thru_rescue_mod = true}
+    parsed = parse('1 rescue 2', :on_rescue_mod) {thru_rescue_mod = true}
     assert_equal true, thru_rescue_mod
+    bug4716 = '[ruby-core:36248]'
+    assert_equal "[rescue_mod(1,2)]", parsed, bug4716
   end
 
   def test_rest_param
@@ -781,6 +1120,15 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     assert_equal true, thru_rest_param
   end
 
+  def test_kwrest_param
+    thru_kwrest = false
+    parse('def a(**) end', :on_kwrest_param) {|n, val| thru_kwrest = val}
+    assert_equal nil, thru_kwrest
+    thru_kwrest = false
+    parse('def a(**x) end', :on_kwrest_param) {|n, val| thru_kwrest = val}
+    assert_equal "x", thru_kwrest
+  end
+
   def test_retry
     thru_retry = false
     parse('retry', :on_retry) {thru_retry = true}
@@ -1044,8 +1392,13 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
 
   def test_words_add
     thru_words_add = false
-    parse('%W[a]', :on_words_add) {thru_words_add = true}
+    tree = parse('%W[a]', :on_words_add) {thru_words_add = true}
     assert_equal true, thru_words_add
+    assert_equal '[array([a])]', tree
+    thru_words_add = false
+    tree = parse('%W[ a ]', :on_words_add) {thru_words_add = true}
+    assert_equal true, thru_words_add
+    assert_equal '[array([a])]', tree
   end
 
   def test_words_new
@@ -1093,7 +1446,6 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
   def test_local_variables
     cmd = 'command(w,[regexp_literal(regexp_add(regexp_new(),25 # ),/)])'
     div = 'binary(ref(w),/,25)'
-    var = '[w]'
     bug1939 = '[ruby-core:24923]'
 
     assert_equal("[#{cmd}]", parse('w /25 # /'), bug1939)
@@ -1106,14 +1458,48 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     assert_equal("[fcall(proc,[],&block([],[void()]))]", parse("proc{|;y|}"))
     if defined?(Process::RLIMIT_AS)
       assert_in_out_err(["-I#{File.dirname(__FILE__)}", "-rdummyparser"],
-                        'Process.setrlimit(Process::RLIMIT_AS,102400); puts DummyParser.new("proc{|;y|}").parse',
-                        ["[fcall(proc,[],&block([],[void()]))]"], [], '[ruby-dev:39423]')
+                        'Process.setrlimit(Process::RLIMIT_AS,100*1024*1024); puts DummyParser.new("proc{|;y|!y}").parse',
+                        ["[fcall(proc,[],&block([],[unary(!,ref(y))]))]"], [], '[ruby-dev:39423]')
     end
   end
 
   def test_unterminated_regexp
-    compile_error = false
-    parse('/', :compile_error) {|e, msg| compile_error = msg}
-    assert_equal("unterminated regexp meets end of file", compile_error)
+    assert_equal("unterminated regexp meets end of file", compile_error('/'))
+  end
+
+  def test_invalid_instance_variable_name
+    assert_equal("`@1' is not allowed as an instance variable name", compile_error('@1'))
+    assert_equal("`@%' is not allowed as an instance variable name", compile_error('@%'))
+    assert_equal("`@' without identifiers is not allowed as an instance variable name", compile_error('@'))
+  end
+
+  def test_invalid_class_variable_name
+    assert_equal("`@@1' is not allowed as a class variable name", compile_error('@@1'))
+    assert_equal("`@@%' is not allowed as a class variable name", compile_error('@@%'))
+    assert_equal("`@@' without identifiers is not allowed as a class variable name", compile_error('@@'))
+  end
+
+  def test_invalid_global_variable_name
+    assert_equal("`$%' is not allowed as a global variable name", compile_error('$%'))
+    assert_equal("`$' without identifiers is not allowed as a global variable name", compile_error('$'))
+  end
+
+  def test_warning_shadowing
+    fmt, *args = warning("x = 1; tap {|;x|}")
+    assert_match(/shadowing outer local variable/, fmt)
+    assert_equal("x", args[0])
+    assert_match(/x/, fmt % args)
+  end
+
+  def test_warning_ignored_magic_comment
+    fmt, *args = warning("1; #-*- frozen-string-literal: true -*-")
+    assert_match(/ignored after any tokens/, fmt)
+    assert_equal("frozen_string_literal", args[0])
+  end
+
+  def test_warn_cr_in_middle
+    fmt = nil
+    assert_warn("") {fmt, = warn("\r;")}
+    assert_match(/encountered/, fmt)
   end
 end if ripper_test
