@@ -1,4 +1,4 @@
-# frozen_string_literal: false
+# frozen_string_literal: true
 #
 # test_scanner_events.rb
 #
@@ -48,70 +48,70 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
   def test_lex
     assert_equal [],
                  Ripper.lex('')
-    assert_equal [[[1,0], :on_ident, "a"]],
+    assert_equal [[[1,0], :on_ident, "a", Ripper::EXPR_CMDARG]],
                  Ripper.lex('a')
-    assert_equal [[[1, 0], :on_kw, "nil"]],
+    assert_equal [[[1, 0], :on_kw, "nil", Ripper::EXPR_END]],
                  Ripper.lex("nil")
-    assert_equal [[[1, 0], :on_kw, "def"],
-                  [[1, 3], :on_sp, " "],
-                  [[1, 4], :on_ident, "m"],
-                  [[1, 5], :on_lparen, "("],
-                  [[1, 6], :on_ident, "a"],
-                  [[1, 7], :on_rparen, ")"],
-                  [[1, 8], :on_kw, "end"]],
+    assert_equal [[[1, 0], :on_kw, "def", Ripper::EXPR_FNAME],
+                  [[1, 3], :on_sp, " ", Ripper::EXPR_FNAME],
+                  [[1, 4], :on_ident, "m", Ripper::EXPR_ENDFN],
+                  [[1, 5], :on_lparen, "(", Ripper::EXPR_BEG | Ripper::EXPR_LABEL],
+                  [[1, 6], :on_ident, "a", Ripper::EXPR_ARG],
+                  [[1, 7], :on_rparen, ")", Ripper::EXPR_ENDFN],
+                  [[1, 8], :on_kw, "end", Ripper::EXPR_END]],
                  Ripper.lex("def m(a)end")
-    assert_equal [[[1, 0], :on_int, "1"],
-                  [[1, 1], :on_nl, "\n"],
-                  [[2, 0], :on_int, "2"],
-                  [[2, 1], :on_nl, "\n"],
-                  [[3, 0], :on_int, "3"]],
+    assert_equal [[[1, 0], :on_int, "1", Ripper::EXPR_END | Ripper::EXPR_ENDARG],
+                  [[1, 1], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[2, 0], :on_int, "2", Ripper::EXPR_END | Ripper::EXPR_ENDARG],
+                  [[2, 1], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[3, 0], :on_int, "3", Ripper::EXPR_END | Ripper::EXPR_ENDARG]],
                  Ripper.lex("1\n2\n3")
-    assert_equal [[[1, 0], :on_heredoc_beg, "<<""EOS"],
-                  [[1, 5], :on_nl, "\n"],
-                  [[2, 0], :on_tstring_content, "heredoc\n"],
-                  [[3, 0], :on_heredoc_end, "EOS"]],
+    assert_equal [[[1, 0], :on_heredoc_beg, "<<""EOS", Ripper::EXPR_BEG],
+                  [[1, 5], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[2, 0], :on_tstring_content, "heredoc\n", Ripper::EXPR_BEG],
+                  [[3, 0], :on_heredoc_end, "EOS", Ripper::EXPR_BEG]],
                  Ripper.lex("<<""EOS\nheredoc\nEOS")
-    assert_equal [[[1, 0], :on_heredoc_beg, "<<""EOS"],
-                  [[1, 5], :on_nl, "\n"],
-                  [[2, 0], :on_heredoc_end, "EOS"]],
+    assert_equal [[[1, 0], :on_heredoc_beg, "<<""EOS", Ripper::EXPR_BEG],
+                  [[1, 5], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[2, 0], :on_heredoc_end, "EOS", Ripper::EXPR_BEG]],
                  Ripper.lex("<<""EOS\nEOS"),
                  "bug#4543"
-    assert_equal [[[1, 0], :on_regexp_beg, "/"],
-                  [[1, 1], :on_tstring_content, "foo\nbar"],
-                  [[2, 3], :on_regexp_end, "/"]],
+    assert_equal [[[1, 0], :on_regexp_beg, "/", Ripper::EXPR_BEG],
+                  [[1, 1], :on_tstring_content, "foo\nbar", Ripper::EXPR_BEG],
+                  [[2, 3], :on_regexp_end, "/", Ripper::EXPR_BEG]],
                  Ripper.lex("/foo\nbar/")
-    assert_equal [[[1, 0], :on_regexp_beg, "/"],
-                  [[1, 1], :on_tstring_content, "foo\n\u3020"],
-                  [[2, 3], :on_regexp_end, "/"]],
+    assert_equal [[[1, 0], :on_regexp_beg, "/", Ripper::EXPR_BEG],
+                  [[1, 1], :on_tstring_content, "foo\n\u3020", Ripper::EXPR_BEG],
+                  [[2, 3], :on_regexp_end, "/", Ripper::EXPR_BEG]],
                  Ripper.lex("/foo\n\u3020/")
-    assert_equal [[[1, 0], :on_tstring_beg, "'"],
-                  [[1, 1], :on_tstring_content, "foo\n\xe3\x80\xa0"],
-                  [[2, 3], :on_tstring_end, "'"]],
+    assert_equal [[[1, 0], :on_tstring_beg, "'", Ripper::EXPR_BEG],
+                  [[1, 1], :on_tstring_content, "foo\n\xe3\x80\xa0", Ripper::EXPR_BEG],
+                  [[2, 3], :on_tstring_end, "'", Ripper::EXPR_END | Ripper::EXPR_ENDARG]],
                  Ripper.lex("'foo\n\xe3\x80\xa0'")
-    assert_equal [[[1, 0], :on_tstring_beg, "'"],
-                  [[1, 1], :on_tstring_content, "\u3042\n\u3044"],
-                  [[2, 3], :on_tstring_end, "'"]],
+    assert_equal [[[1, 0], :on_tstring_beg, "'", Ripper::EXPR_BEG],
+                  [[1, 1], :on_tstring_content, "\u3042\n\u3044", Ripper::EXPR_BEG],
+                  [[2, 3], :on_tstring_end, "'", Ripper::EXPR_END | Ripper::EXPR_ENDARG]],
                  Ripper.lex("'\u3042\n\u3044'")
-    assert_equal [[[1, 0], :on_rational, "1r"],
-                  [[1, 2], :on_nl, "\n"],
-                  [[2, 0], :on_imaginary, "2i"],
-                  [[2, 2], :on_nl, "\n"],
-                  [[3, 0], :on_imaginary, "3ri"],
-                  [[3, 3], :on_nl, "\n"],
-                  [[4, 0], :on_rational, "4.2r"],
-                  [[4, 4], :on_nl, "\n"],
-                  [[5, 0], :on_imaginary, "5.6ri"],
+    assert_equal [[[1, 0], :on_rational, "1r", Ripper::EXPR_END | Ripper::EXPR_ENDARG],
+                  [[1, 2], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[2, 0], :on_imaginary, "2i", Ripper::EXPR_END | Ripper::EXPR_ENDARG],
+                  [[2, 2], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[3, 0], :on_imaginary, "3ri", Ripper::EXPR_END | Ripper::EXPR_ENDARG],
+                  [[3, 3], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[4, 0], :on_rational, "4.2r", Ripper::EXPR_END | Ripper::EXPR_ENDARG],
+                  [[4, 4], :on_nl, "\n", Ripper::EXPR_BEG],
+                  [[5, 0], :on_imaginary, "5.6ri", Ripper::EXPR_END | Ripper::EXPR_ENDARG],
                  ],
                  Ripper.lex("1r\n2i\n3ri\n4.2r\n5.6ri")
-     assert_equal [[[1, 0], :on_heredoc_beg, "<<~EOS"],
-                   [[1, 6], :on_nl, "\n"],
-                   [[2, 0], :on_ignored_sp, "  "],
-                   [[2, 2], :on_tstring_content, "heredoc\n"],
-                   [[3, 0], :on_heredoc_end, "EOS"]
+     assert_equal [[[1, 0], :on_heredoc_beg, "<<~EOS", Ripper::EXPR_BEG],
+                   [[1, 6], :on_nl, "\n", Ripper::EXPR_BEG],
+                   [[2, 0], :on_ignored_sp, "  ", Ripper::EXPR_BEG],
+                   [[2, 2], :on_tstring_content, "heredoc\n", Ripper::EXPR_BEG],
+                   [[3, 0], :on_heredoc_end, "EOS", Ripper::EXPR_BEG]
                  ],
                  Ripper.lex("<<~EOS\n  heredoc\nEOS")
-    assert_equal [[[1, 0], :on_tstring_beg, "'"],
-                  [[1, 1], :on_tstring_content, "foo"]],
+    assert_equal [[[1, 0], :on_tstring_beg, "'", Ripper::EXPR_BEG],
+                  [[1, 1], :on_tstring_content, "foo", Ripper::EXPR_BEG]],
                  Ripper.lex("'foo")
   end
 
@@ -140,7 +140,7 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
   end
 
   def assert_location(src)
-    buf = ''
+    buf = ''.dup
     Ripper.lex(src).each do |pos, type, tok|
       line, col = *pos
       assert_equal buf.count("\n") + 1, line,
@@ -637,8 +637,10 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('words_beg', '%W()')
     assert_equal ['%W('],
                  scan('words_beg', '%W(w w w)')
-    assert_equal ['%W( '],
+    assert_equal ['%W('],
                  scan('words_beg', '%W( w w w )')
+    assert_equal ['%W('],
+                 scan('words_beg', "%W(\nw)")
   end
 
   def test_qwords_beg
@@ -648,8 +650,10 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('qwords_beg', '%w()')
     assert_equal ['%w('],
                  scan('qwords_beg', '%w(w w w)')
-    assert_equal ['%w( '],
+    assert_equal ['%w('],
                  scan('qwords_beg', '%w( w w w )')
+    assert_equal ['%w('],
+                 scan('qwords_beg', "%w(\nw)")
   end
 
   def test_qsymbols_beg
@@ -659,8 +663,10 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('qsymbols_beg', '%i()')
     assert_equal ['%i('],
                  scan('qsymbols_beg', '%i(w w w)')
-    assert_equal ['%i( '],
+    assert_equal ['%i('],
                  scan('qsymbols_beg', '%i( w w w )')
+    assert_equal ['%i('],
+                 scan('qsymbols_beg', "%i(\nw)")
   end
 
   def test_symbols_beg
@@ -670,22 +676,25 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('symbols_beg', '%I()')
     assert_equal ['%I('],
                  scan('symbols_beg', '%I(w w w)')
-    assert_equal ['%I( '],
+    assert_equal ['%I('],
                  scan('symbols_beg', '%I( w w w )')
+    assert_equal ['%I('],
+                 scan('symbols_beg', "%I(\nw)")
   end
 
-  # FIXME: Close paren must not present (`words_end' scanner event?).
   def test_words_sep
     assert_equal [],
                  scan('words_sep', '')
-    assert_equal [')'],
+    assert_equal [],
                  scan('words_sep', '%w()')
-    assert_equal [' ', ' ', ')'],
+    assert_equal [' ', ' '],
                  scan('words_sep', '%w(w w w)')
-    assert_equal [' ', ' ', ' )'],
+    assert_equal [' ', ' ', ' ', ' '],
                  scan('words_sep', '%w( w w w )')
-    assert_equal ["\n", ' ', ' )'],
+    assert_equal [' ', "\n", ' ', ' '],
                  scan('words_sep', "%w( w\nw w )")
+    assert_equal ["\n\n", "\n ", ' ', ' '],
+                 scan('words_sep', "%w(\n\nw\n w w )")
   end
 
   def test_heredoc_beg
