@@ -1,3 +1,4 @@
+# frozen_string_literal: false
 require 'test/unit'
 
 class Array
@@ -5,8 +6,8 @@ class Array
     collect{|e| [e, yield(e)]}.sort{|a,b|a[1]<=>b[1]}
   end
   def iter_test2
-    a = collect{|e| [e, yield(e)]}
-    a.sort{|a,b|a[1]<=>b[1]}
+    ary = collect{|e| [e, yield(e)]}
+    ary.sort{|a,b|a[1]<=>b[1]}
   end
 end
 
@@ -25,14 +26,14 @@ class TestIterator < Test::Unit::TestCase
   end
 
   def test_array
-    $x = [1, 2, 3, 4]
-    $y = []
+    x = [1, 2, 3, 4]
+    y = []
 
     # iterator over array
-    for i in $x
-      $y.push i
+    for i in x
+      y.push i
     end
-    assert_equal($x, $y)
+    assert_equal(x, y)
   end
 
   def tt
@@ -51,7 +52,7 @@ class TestIterator < Test::Unit::TestCase
 
   def test_nested_iterator
     i = 0
-    tt{|i| break if i == 5}
+    tt{|j| break if j == 5}
     assert_equal(0, i)
 
     assert_raise(ArgumentError) do
@@ -73,42 +74,42 @@ class TestIterator < Test::Unit::TestCase
   def test_break
     done = true
     loop{
-      break
+      break if true
       done = false			# should not reach here
     }
     assert(done)
 
     done = false
-    $bad = false
+    bad = false
     loop {
       break if done
       done = true
-      next
-      $bad = true			# should not reach here
+      next if true
+      bad = true			# should not reach here
     }
-    assert(!$bad)
+    assert(!bad)
 
     done = false
-    $bad = false
+    bad = false
     loop {
       break if done
       done = true
-      redo
-      $bad = true			# should not reach here
+      redo if true
+      bad = true			# should not reach here
     }
-    assert(!$bad)
+    assert(!bad)
 
-    $x = []
+    x = []
     for i in 1 .. 7
-      $x.push i
+      x.push i
     end
-    assert_equal(7, $x.size)
-    assert_equal([1, 2, 3, 4, 5, 6, 7], $x)
+    assert_equal(7, x.size)
+    assert_equal([1, 2, 3, 4, 5, 6, 7], x)
   end
 
   def test_append_method_to_built_in_class
-    $x = [[1,2],[3,4],[5,6]]
-    assert_equal($x.iter_test1{|x|x}, $x.iter_test2{|x|x})
+    x = [[1,2],[3,4],[5,6]]
+    assert_equal(x.iter_test1{|e|e}, x.iter_test2{|e|e})
   end
 
   class IterTest
@@ -278,6 +279,9 @@ class TestIterator < Test::Unit::TestCase
   def proc_call(&b)
     b.call
   end
+  def proc_call2(b)
+    b.call
+  end
   def proc_yield()
     yield
   end
@@ -299,6 +303,7 @@ class TestIterator < Test::Unit::TestCase
 
   def test_ljump
     assert_raise(LocalJumpError) {get_block{break}.call}
+    assert_raise(LocalJumpError) {proc_call2(get_block{break}){}}
 
     # cannot use assert_nothing_raised due to passing block.
     begin
@@ -306,7 +311,7 @@ class TestIterator < Test::Unit::TestCase
     rescue LocalJumpError
       assert(false, "LocalJumpError occurred from break in lambda")
     else
-      assert(11, val)
+      assert_equal(11, val)
     end
 
     block = get_block{11}
