@@ -1,5 +1,5 @@
+# frozen_string_literal: false
 require 'test/unit'
-require_relative 'envutil'
 
 class TestKeywordArguments < Test::Unit::TestCase
   def f1(str: "foo", num: 424242)
@@ -311,11 +311,16 @@ class TestKeywordArguments < Test::Unit::TestCase
     feature7701 = '[ruby-core:51454] [Feature #7701] required keyword argument'
     o = Object.new
     assert_nothing_raised(SyntaxError, feature7701) do
-      eval("def o.foo(a:) a; end")
+      eval("def o.foo(a:) a; end", nil, "xyzzy")
       eval("def o.bar(a:,**b) [a, b]; end")
     end
     assert_raise_with_message(ArgumentError, /missing keyword/, feature7701) {o.foo}
     assert_raise_with_message(ArgumentError, /unknown keyword/, feature7701) {o.foo(a:0, b:1)}
+    begin
+      o.foo(a: 0, b: 1)
+    rescue => e
+      assert_equal('xyzzy', e.backtrace_locations[0].path)
+    end
     assert_equal(42, o.foo(a: 42), feature7701)
     assert_equal([[:keyreq, :a]], o.method(:foo).parameters, feature7701)
 
@@ -363,22 +368,33 @@ class TestKeywordArguments < Test::Unit::TestCase
   def test_block_required_keyword
     feature7701 = '[ruby-core:51454] [Feature #7701] required keyword argument'
     b = assert_nothing_raised(SyntaxError, feature7701) do
-      break eval("proc {|a:| a}")
+      break eval("proc {|a:| a}", nil, 'xyzzy', __LINE__)
     end
     assert_raise_with_message(ArgumentError, /missing keyword/, feature7701) {b.call}
-    assert_raise_with_message(ArgumentError, /unknown keyword/, feature7701) {b.call(a:0, b:1)}
+    e = assert_raise_with_message(ArgumentError, /unknown keyword/, feature7701) {b.call(a:0, b:1)}
+    assert_equal('xyzzy', e.backtrace_locations[0].path)
+
     assert_equal(42, b.call(a: 42), feature7701)
     assert_equal([[:keyreq, :a]], b.parameters, feature7701)
 
     bug8139 = '[ruby-core:53608] [Bug #8139] required keyword argument with rest hash'
     b = assert_nothing_raised(SyntaxError, feature7701) do
-      break eval("proc {|a:, **b| [a, b]}")
+      break eval("proc {|a:, **bl| [a, bl]}", nil, __FILE__, __LINE__)
     end
     assert_equal([42, {}], b.call(a: 42), feature7701)
     assert_equal([42, {c: feature7701}], b.call(a: 42, c: feature7701), feature7701)
-    assert_equal([[:keyreq, :a], [:keyrest, :b]], b.parameters, feature7701)
+    assert_equal([[:keyreq, :a], [:keyrest, :bl]], b.parameters, feature7701)
     assert_raise_with_message(ArgumentError, /missing keyword/, bug8139) {b.call(c: bug8139)}
     assert_raise_with_message(ArgumentError, /missing keyword/, bug8139) {b.call}
+
+    b = assert_nothing_raised(SyntaxError, feature7701) do
+      break eval("proc {|m, a:| [m, a]}", nil, 'xyzzy', __LINE__)
+    end
+    assert_raise_with_message(ArgumentError, /missing keyword/) {b.call}
+    assert_equal([:ok, 42], b.call(:ok, a: 42))
+    e = assert_raise_with_message(ArgumentError, /unknown keyword/) {b.call(42, a:0, b:1)}
+    assert_equal('xyzzy', e.backtrace_locations[0].path)
+    assert_equal([[:opt, :m], [:keyreq, :a]], b.parameters)
   end
 
   def test_super_with_keyword
@@ -440,7 +456,7 @@ class TestKeywordArguments < Test::Unit::TestCase
       end
     end
     assert_equal([{}, {}], a.new.foo({}))
-    assert_equal([{}, {:bar=>"x"}], a.new.foo({}, bar: "x"))
+    assert_equal([{}, {:bar=>"x"}], a.new.foo({}, bar: "x"), bug8040)
   end
 
   def test_precedence_of_keyword_arguments_with_post_argument
@@ -487,6 +503,29 @@ class TestKeywordArguments < Test::Unit::TestCase
     assert_equal([1, 9], m1(1, o, &->(a, k: 0) {break [a, k]}), bug10016)
   end
 
+  def test_splat_hash
+    m = Object.new
+    def m.f() :ok; end
+    def m.f2(a = nil) a; end
+    o = {a: 1}
+    assert_raise_with_message(ArgumentError, /unknown keyword: a/) {
+      m.f(**o)
+    }
+    o = {}
+    assert_equal(:ok, m.f(**o), '[ruby-core:68124] [Bug #10856]')
+    a = []
+    assert_equal(:ok, m.f(*a, **o), '[ruby-core:83638] [Bug #10856]')
+
+    o = {a: 42}
+    assert_equal({a: 42}, m.f2(**o), '[ruby-core:82280] [Bug #13791]')
+
+    assert_equal({a: 42}, m.f2("a".to_sym => 42), '[ruby-core:82291] [Bug #13793]')
+
+    o = {}
+    a = [:ok]
+    assert_equal(:ok, m.f2(*a, **o), '[ruby-core:83638] [Bug #10856]')
+  end
+
   def test_gced_object_in_stack
     bug8964 = '[ruby-dev:47729] [Bug #8964]'
     assert_normal_exit %q{
@@ -506,6 +545,15 @@ class TestKeywordArguments < Test::Unit::TestCase
     }, bug8964
   end
 
+  def test_dynamic_symbol_keyword
+    bug10266 = '[ruby-dev:48564] [Bug #10266]'
+    assert_separately(['-', bug10266], <<-'end;') #    do
+      bug = ARGV.shift
+      "hoge".to_sym
+      assert_nothing_raised(bug) {eval("def a(hoge:); end")}
+    end;
+  end
+
   def test_unknown_keyword_with_block
     bug10413 = '[ruby-core:65837] [Bug #10413]'
     class << (o = Object.new)
@@ -520,4 +568,113 @@ class TestKeywordArguments < Test::Unit::TestCase
       o.foo {raise "unreachable"}
     }
   end
+
+  def test_unknown_keyword
+    bug13004 = '[ruby-dev:49893] [Bug #13004]'
+    assert_raise_with_message(ArgumentError, /unknown keyword: invalid-argument/, bug13004) {
+      [].sample(random: nil, "invalid-argument": nil)
+    }
+  end
+
+  def test_super_with_anon_restkeywords
+    bug10659 = '[ruby-core:67157] [Bug #10659]'
+
+    foo = Class.new do
+      def foo(**h)
+        h
+      end
+    end
+
+    class << (obj = foo.new)
+      def foo(bar: "bar", **)
+        super
+      end
+    end
+
+    assert_nothing_raised(TypeError, bug10659) {
+      assert_equal({:bar => "bar"}, obj.foo, bug10659)
+    }
+  end
+
+  def m(a) yield a end
+
+  def test_nonsymbol_key
+    result = m(["a" => 10]) { |a = nil, **b| [a, b] }
+    assert_equal([{"a" => 10}, {}], result)
+  end
+
+  def method_for_test_to_hash_call_during_setup_complex_parameters k1:, k2:, **rest_kw
+    [k1, k2, rest_kw]
+  end
+
+  def test_to_hash_call_during_setup_complex_parameters
+    sym = "sym_#{Time.now}".to_sym
+    h = method_for_test_to_hash_call_during_setup_complex_parameters k1: "foo", k2: "bar", sym => "baz"
+    assert_equal ["foo", "bar", {sym => "baz"}], h, '[Bug #11027]'
+  end
+
+  class AttrSetTest
+    attr_accessor :foo
+    alias set_foo :foo=
+  end
+
+  def test_attr_set_method_cache
+    obj = AttrSetTest.new
+    h = {a: 1, b: 2}
+    2.times{
+      obj.foo = 1
+      assert_equal(1, obj.foo)
+      obj.set_foo 2
+      assert_equal(2, obj.foo)
+      obj.set_foo(x: 1, y: 2)
+      assert_equal({x: 1, y: 2}, obj.foo)
+      obj.set_foo(x: 1, y: 2, **h)
+      assert_equal({x: 1, y: 2, **h}, obj.foo)
+    }
+  end
+
+  def test_kwrest_overwritten
+    bug13015 = '[ruby-core:78536] [Bug #13015]'
+
+    klass = EnvUtil.labeled_class("Parent") do
+      def initialize(d:)
+      end
+    end
+
+    klass = EnvUtil.labeled_class("Child", klass) do
+      def initialize(d:, **h)
+        h = [2, 3]
+        super
+      end
+    end
+
+    assert_raise_with_message(TypeError, /expected Hash/, bug13015) do
+      klass.new(d: 4)
+    end
+  end
+
+  def test_non_keyword_hash_subclass
+    bug12884 = '[ruby-core:77813] [Bug #12884]'
+    klass = EnvUtil.labeled_class("Child", Hash)
+    obj = Object.new
+    def obj.t(params = klass.new, d: nil); params; end
+    x = klass.new
+    x["foo"] = "bar"
+    result = obj.t(x)
+    assert_equal(x, result)
+    assert_kind_of(klass, result, bug12884)
+  end
+
+  def test_arity_error_message
+    obj = Object.new
+    def obj.t(x:) end
+    assert_raise_with_message(ArgumentError, /required keyword: x\)/) do
+      obj.t(42)
+    end
+    obj = Object.new
+    def obj.t(x:, y:, z: nil) end
+    assert_raise_with_message(ArgumentError, /required keywords: x, y\)/) do
+      obj.t(42)
+    end
+  end
 end
