@@ -14,7 +14,14 @@ class TestLazyEnumerator < Test::Unit::TestCase
 
     def each(*args)
       @args = args
-      @enum.each {|i| @current = i; yield i}
+      @enum.each do |v|
+        @current = v
+        if v.is_a? Enumerable
+          yield *v
+        else
+          yield v
+        end
+      end
     end
   end
 
@@ -25,7 +32,7 @@ class TestLazyEnumerator < Test::Unit::TestCase
 
     a = [1, 2, 3].lazy
     a.freeze
-    assert_raise(RuntimeError) {
+    assert_raise(FrozenError) {
       a.__send__ :initialize, [4, 5], &->(y, *v) { y << yield(*v) }
     }
   end
@@ -100,6 +107,15 @@ class TestLazyEnumerator < Test::Unit::TestCase
     assert_equal(1, a.current)
   end
 
+  def test_map_packed_nested
+    bug = '[ruby-core:81638] [Bug#13648]'
+
+    a = Step.new([[1, 2]])
+    expected = [[[1, 2]]]
+    assert_equal(expected, a.map {|*args| args}.map {|*args| args}.to_a)
+    assert_equal(expected, a.lazy.map {|*args| args}.map {|*args| args}.to_a, bug)
+  end
+
   def test_flat_map
     a = Step.new(1..3)
     assert_equal(2, a.flat_map {|x| [x * 2]}.first)
@@ -275,6 +291,11 @@ class TestLazyEnumerator < Test::Unit::TestCase
     assert_equal(nil, a.current)
   end
 
+  def test_take_bad_arg
+    a = Step.new(1..10)
+    assert_raise(ArgumentError) { a.lazy.take(-1) }
+  end
+
   def test_take_recycle
     bug6428 = '[ruby-dev:45634]'
     a = Step.new(1..10)
@@ -480,6 +501,15 @@ EOS
     assert_equal Float::INFINITY, loop.lazy.cycle.size
     assert_equal nil, lazy.select{}.cycle(4).size
     assert_equal nil, lazy.select{}.cycle.size
+
+    class << (obj = Object.new)
+      def each; end
+      def size; 0; end
+      include Enumerable
+    end
+    lazy = obj.lazy
+    assert_equal 0, lazy.cycle.size
+    assert_raise(TypeError) {lazy.cycle("").size}
   end
 
   def test_map_zip
