@@ -1,8 +1,8 @@
-require 'test/unit'
-require File.join(File.expand_path(File.dirname(__FILE__)), 'gemutilities')
+# frozen_string_literal: true
+require 'rubygems/test_case'
 require 'rubygems/commands/lock_command'
 
-class TestGemCommandsLockCommand < RubyGemTestCase
+class TestGemCommandsLockCommand < Gem::TestCase
 
   def setup
     super
@@ -46,7 +46,7 @@ gem 'a', '= 1'
     expected = <<-EXPECTED
 require 'rubygems'
 gem 'd', '= 1'
-# Unable to satisfy 'z (>= 0, runtime)' from currently installed gems
+# Unable to satisfy 'z (>= 0)' from currently installed gems
     EXPECTED
 
     assert_equal expected, @ui.output
@@ -56,7 +56,7 @@ gem 'd', '= 1'
   def test_execute_strict
     @cmd.handle_options %w[c-1 --strict]
 
-    e = assert_raise Gem::Exception do
+    e = assert_raises Gem::Exception do
       use_ui @ui do
         @cmd.execute
       end
