@@ -1,3 +1,4 @@
+# frozen_string_literal: true
 require 'rubygems/test_case'
 require 'rubygems/spec_fetcher'
 
@@ -13,51 +14,49 @@ class TestGemSpecFetcher < Gem::TestCase
     @uri = URI.parse @gem_repo
     @source = Gem::Source.new(@uri)
 
-    util_setup_fake_fetcher
-
-    @a_pre = new_spec 'a', '1.a'
-
-    install_specs @a_pre
-
-    Gem::Specification.remove_spec @b2
-
-    all = Gem::Specification.map { |spec|
-      Gem::NameTuple.new(spec.name, spec.version, spec.original_platform)
-    }.sort
-
-    @prerelease_specs, @specs = all.partition { |g| g.prerelease? }
-
-    # TODO: couldn't all of this come from the fake spec fetcher?
-    @latest_specs = Gem::Specification.latest_specs.sort.map { |spec|
-      Gem::NameTuple.new(spec.name, spec.version, spec.original_platform)
-    }
-
-    v = Gem.marshal_version
-    s_zip = util_gzip(Marshal.dump(Gem::NameTuple.to_basic(@specs)))
-    l_zip = util_gzip(Marshal.dump(Gem::NameTuple.to_basic(@latest_specs)))
-    p_zip = util_gzip(Marshal.dump(Gem::NameTuple.to_basic(@prerelease_specs)))
-    @fetcher.data["#{@gem_repo}specs.#{v}.gz"]            = s_zip
-    @fetcher.data["#{@gem_repo}latest_specs.#{v}.gz"]     = l_zip
-    @fetcher.data["#{@gem_repo}prerelease_specs.#{v}.gz"] = p_zip
-
     @sf = Gem::SpecFetcher.new
+  end
 
-    @released = Gem::NameTuple.from_list \
-                 [["a",      Gem::Version.new("1"),   "ruby"],
-                  ["a",      Gem::Version.new("2"),   "ruby"],
-                  ["a_evil", Gem::Version.new("9"),   "ruby"],
-                  ["c",      Gem::Version.new("1.2"), "ruby"],
-                  ['dep_x',  Gem::Version.new(1),     'ruby'],
-                  ["pl",     Gem::Version.new("1"),   "i386-linux"],
-                  ['x',  Gem::Version.new(1),     'ruby']]
+  def test_initialize
+    fetcher = Gem::SpecFetcher.new
+
+    assert_same Gem.sources, fetcher.sources
+  end
+
+  def test_initialize_source
+    alternate = 'http://alternate.example'
+    fetcher = Gem::SpecFetcher.new alternate
+
+    refute_same Gem.sources, fetcher.sources
+
+    assert_equal alternate, fetcher.sources
+  end
+
+  def test_initialize_nonexistent_home_dir
+    FileUtils.rmdir Gem.user_home
+
+    assert Gem::SpecFetcher.new
+  end
+
+  def test_initialize_unwritable_home_dir
+    skip 'chmod not supported' if Gem.win_platform?
+
+    FileUtils.chmod 0000, Gem.user_home
+
+    begin
+      assert Gem::SpecFetcher.new
+    ensure
+      FileUtils.chmod 0755, Gem.user_home
+    end
   end
 
   def test_spec_for_dependency_all
-    d = "#{@gem_repo}#{Gem::MARSHAL_SPEC_DIR}"
-    @fetcher.data["#{d}#{@a1.spec_name}.rz"]    = util_zip(Marshal.dump(@a1))
-    @fetcher.data["#{d}#{@a2.spec_name}.rz"]    = util_zip(Marshal.dump(@a2))
-    @fetcher.data["#{d}#{@a_pre.spec_name}.rz"] = util_zip(Marshal.dump(@a_pre))
-    @fetcher.data["#{d}#{@a3a.spec_name}.rz"]   = util_zip(Marshal.dump(@a3a))
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', 1
+      fetcher.spec 'a', '2.a'
+      fetcher.spec 'a', 2
+      fetcher.spec 'a', '3.a'
+    end
 
     dep = Gem::Dependency.new 'a', ">= 1"
 
@@ -67,7 +66,7 @@ class TestGemSpecFetcher < Gem::TestCase
       [spec.full_name, source_uri]
     end
 
-    expected = [[@a1.full_name, @source], [@a2.full_name, @source]]
+    expected = [['a-1', @source], ['a-2', @source]]
 
     assert_equal expected, spec_names
 
@@ -75,10 +74,11 @@ class TestGemSpecFetcher < Gem::TestCase
   end
 
   def test_spec_for_dependency_latest
-    d = "#{@gem_repo}#{Gem::MARSHAL_SPEC_DIR}"
-    @fetcher.data["#{d}#{@a1.spec_name}.rz"]    = util_zip(Marshal.dump(@a1))
-    @fetcher.data["#{d}#{@a2.spec_name}.rz"]    = util_zip(Marshal.dump(@a2))
-    @fetcher.data["#{d}#{@a_pre.spec_name}.rz"] = util_zip(Marshal.dump(@a_pre))
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', 1
+      fetcher.spec 'a', 2
+      fetcher.spec 'a', '3.a'
+    end
 
     dep = Gem::Dependency.new 'a'
     specs_and_sources, _ = @sf.spec_for_dependency dep
@@ -87,14 +87,15 @@ class TestGemSpecFetcher < Gem::TestCase
       [spec.full_name, source_uri]
     end
 
-    assert_equal [[@a2.full_name, Gem::Source.new(@gem_repo)]], spec_names
+    assert_equal [['a-2', Gem::Source.new(@gem_repo)]],
+                 spec_names
   end
 
   def test_spec_for_dependency_prerelease
-    d = "#{@gem_repo}#{Gem::MARSHAL_SPEC_DIR}"
-    @fetcher.data["#{d}#{@a1.spec_name}.rz"]    = util_zip(Marshal.dump(@a1))
-    @fetcher.data["#{d}#{@a2.spec_name}.rz"]    = util_zip(Marshal.dump(@a2))
-    @fetcher.data["#{d}#{@a_pre.spec_name}.rz"] = util_zip(Marshal.dump(@a_pre))
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', '1.a'
+      fetcher.spec 'a', 1
+    end
 
     specs_and_sources, _ = @sf.spec_for_dependency dep('a', '1.a')
 
@@ -102,14 +103,15 @@ class TestGemSpecFetcher < Gem::TestCase
       [spec.full_name, source_uri]
     end
 
-    assert_equal [[@a_pre.full_name, Gem::Source.new(@gem_repo)]], spec_names
+    assert_equal [['a-1.a', Gem::Source.new(@gem_repo)]], spec_names
   end
 
   def test_spec_for_dependency_platform
     util_set_arch 'i386-linux'
 
-    @fetcher.data["#{@gem_repo}#{Gem::MARSHAL_SPEC_DIR}#{@pl1.original_name}.gemspec.rz"] =
-      util_zip(Marshal.dump(@pl1))
+    spec_fetcher do |fetcher|
+      fetcher.legacy_platform
+    end
 
     dep = Gem::Dependency.new 'pl', 1
     specs_and_sources, _ = @sf.spec_for_dependency dep
@@ -118,22 +120,26 @@ class TestGemSpecFetcher < Gem::TestCase
       [spec.full_name, source_uri]
     end
 
-    assert_equal [[@pl1.full_name, Gem::Source.new(@gem_repo)]], spec_names
+    assert_equal [['pl-1-x86-linux', Gem::Source.new(@gem_repo)]],
+                 spec_names
   end
 
   def test_spec_for_dependency_mismatched_platform
     util_set_arch 'hrpa-989'
 
-    @fetcher.data["#{@gem_repo}#{Gem::MARSHAL_SPEC_DIR}#{@pl1.original_name}.gemspec.rz"] =
-      util_zip(Marshal.dump(@pl1))
+    spec_fetcher do |fetcher|
+      fetcher.legacy_platform
+    end
 
     dep = Gem::Dependency.new 'pl', 1
     specs_and_sources, errors = @sf.spec_for_dependency dep
 
     assert_equal 0, specs_and_sources.size
     assert_equal 1, errors.size
+    pmm = errors.first
 
-    assert_equal "i386-linux", errors[0].platforms.first
+    assert_equal "i386-linux", pmm.platforms.first
+    assert_equal "Found pl (1), but was for platform i386-linux", pmm.wordy
   end
 
   def test_spec_for_dependency_bad_fetch_spec
@@ -144,11 +150,12 @@ class TestGemSpecFetcher < Gem::TestCase
 
     Gem.sources.replace [src]
 
-    d = "#{@gem_repo}#{Gem::MARSHAL_SPEC_DIR}"
-    @fetcher.data["#{d}#{@a1.spec_name}.rz"]    = util_zip(Marshal.dump(@a1))
-    @fetcher.data["#{d}#{@a2.spec_name}.rz"]    = util_zip(Marshal.dump(@a2))
-    @fetcher.data["#{d}#{@a_pre.spec_name}.rz"] = util_zip(Marshal.dump(@a_pre))
-    @fetcher.data["#{d}#{@a3a.spec_name}.rz"]   = util_zip(Marshal.dump(@a3a))
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', 1
+      fetcher.spec 'a', '2.a'
+      fetcher.spec 'a', 2
+      fetcher.spec 'a', '3.a'
+    end
 
     dep = Gem::Dependency.new 'a', ">= 1"
 
@@ -162,32 +169,91 @@ class TestGemSpecFetcher < Gem::TestCase
     assert_equal "bad news from the internet (#{@gem_repo})", sfp.error.message
   end
 
+  def test_suggest_gems_from_name_latest
+    spec_fetcher do|fetcher|
+      fetcher.spec 'example', 1
+      fetcher.spec 'other-example', 1
+    end
+
+    suggestions = @sf.suggest_gems_from_name('examplw')
+    assert_equal ['example'], suggestions
+  end
+
+  def test_suggest_gems_from_name_prerelease
+    spec_fetcher do|fetcher|
+      fetcher.spec 'example', '1.a'
+      fetcher.spec 'other-example', 1
+    end
+
+    suggestions = @sf.suggest_gems_from_name('examplw')
+    assert_equal ['example'], suggestions
+  end
+
   def test_available_specs_latest
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', 1
+      fetcher.spec 'a', 2
+      fetcher.spec 'a', '3.a'
+      fetcher.legacy_platform
+    end
+
     specs, _ = @sf.available_specs(:latest)
 
     assert_equal [@source], specs.keys
-    assert_equal @latest_specs, specs[@source].sort
+
+    expected = Gem::NameTuple.from_list \
+      [['a',      v(2),     Gem::Platform::RUBY],
+       ['pl',     v(1),     'i386-linux']]
+
+    assert_equal expected, specs[@source]
   end
 
   def test_available_specs_released
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', 1
+      fetcher.legacy_platform
+    end
+
     specs, _ = @sf.available_specs(:released)
 
     assert_equal [@source], specs.keys
 
-    assert_equal @released, specs[@source].sort
+    expected = Gem::NameTuple.from_list \
+      [['a',      v(1),     Gem::Platform::RUBY],
+       ['pl',     v(1),     'i386-linux']]
+
+    assert_equal expected, specs[@source]
   end
 
   def test_available_specs_complete
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', 1
+      fetcher.spec 'a', '2.a'
+      fetcher.spec 'b', 2
+      fetcher.legacy_platform
+    end
+
     specs, _ = @sf.available_specs(:complete)
 
     assert_equal [@source], specs.keys
 
-    comp = @prerelease_specs + @released
+    expected = Gem::NameTuple.from_list \
+      [['a',      v(1),     Gem::Platform::RUBY],
+       ['a',      v('2.a'), Gem::Platform::RUBY],
+       ['b',      v(2),     Gem::Platform::RUBY],
+       ['pl',     v(1),     'i386-linux']]
 
-    assert_equal comp.sort, specs[@source].sort
+    assert_equal expected, specs[@source]
   end
 
   def test_available_specs_complete_handles_no_prerelease
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', 1
+      fetcher.spec 'a', '2.a'
+      fetcher.spec 'b', 2
+      fetcher.legacy_platform
+    end
+
     v = Gem.marshal_version
     @fetcher.data.delete "#{@gem_repo}prerelease_specs.#{v}.gz"
 
@@ -195,13 +261,19 @@ class TestGemSpecFetcher < Gem::TestCase
 
     assert_equal [@source], specs.keys
 
-    comp = @released
+    expected = Gem::NameTuple.from_list \
+      [['a',      v(1), Gem::Platform::RUBY],
+       ['b',      v(2), Gem::Platform::RUBY],
+       ['pl',     v(1), 'i386-linux']]
 
-    assert_equal comp.sort, specs[@source].sort
+    assert_equal expected, specs[@source]
   end
 
-
   def test_available_specs_cache
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', 1
+    end
+
     specs, _ = @sf.available_specs(:latest)
 
     refute specs[@source].empty?
@@ -214,6 +286,13 @@ class TestGemSpecFetcher < Gem::TestCase
   end
 
   def test_available_specs_cache_released
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', 1
+      fetcher.spec 'a', '2.a'
+      fetcher.spec 'b', 2
+      fetcher.legacy_platform
+    end
+
     specs, _ = @sf.available_specs(:released)
 
     refute specs[@source].empty?
@@ -226,9 +305,17 @@ class TestGemSpecFetcher < Gem::TestCase
   end
 
   def test_available_specs_prerelease
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', 1
+      fetcher.spec 'a', '2.a'
+    end
+
     specs, _ = @sf.available_specs(:prerelease)
 
-    assert_equal @prerelease_specs, specs[@source].sort
+    expected = Gem::NameTuple.from_list \
+      [['a',  v('2.a'), Gem::Platform::RUBY]]
+
+    assert_equal expected, specs[@source]
   end
 
   def test_available_specs_with_bad_source
