@@ -1,3 +1,5 @@
+# frozen_string_literal: true
+
 require 'time'
 require 'test/unit'
 
@@ -97,23 +99,23 @@ class TestTimeExtension < Test::Unit::TestCase # :nodoc:
   def subtest_xmlschema_alias(method)
     t = Time.utc(1985, 4, 12, 23, 20, 50, 520000)
     s = "1985-04-12T23:20:50.52Z"
-    assert_equal(t, Time.iso8601(s))
-    assert_equal(s, t.iso8601(2))
+    assert_equal(t, Time.__send__(method, s))
+    assert_equal(s, t.__send__(method, 2))
 
     t = Time.utc(1996, 12, 20, 0, 39, 57)
     s = "1996-12-19T16:39:57-08:00"
-    assert_equal(t, Time.iso8601(s))
+    assert_equal(t, Time.__send__(method, s))
     # There is no way to generate time string with arbitrary timezone.
     s = "1996-12-20T00:39:57Z"
-    assert_equal(t, Time.iso8601(s))
+    assert_equal(t, Time.__send__(method, s))
     assert_equal(s, t.iso8601)
 
     t = Time.utc(1990, 12, 31, 23, 59, 60)
     s = "1990-12-31T23:59:60Z"
-    assert_equal(t, Time.iso8601(s))
+    assert_equal(t, Time.__send__(method, s))
     # leap second is representable only if timezone file has it.
     s = "1990-12-31T15:59:60-08:00"
-    assert_equal(t, Time.iso8601(s))
+    assert_equal(t, Time.__send__(method, s))
 
     begin
       Time.at(-1)
@@ -122,7 +124,7 @@ class TestTimeExtension < Test::Unit::TestCase # :nodoc:
     else
       t = Time.utc(1937, 1, 1, 11, 40, 27, 870000)
       s = "1937-01-01T12:00:27.87+00:20"
-      assert_equal(t, Time.iso8601(s))
+      assert_equal(t, Time.__send__(method, s))
     end
   end
 
@@ -336,6 +338,18 @@ class TestTimeExtension < Test::Unit::TestCase # :nodoc:
     assert_equal(t1.utc?, t2.utc?)
   end
 
+  def test_parse_now_nil
+    assert_equal(Time.new(2000,1,1,0,0,0,"+11:00"),
+                 Time.parse("2000-01-01T00:00:00+11:00", nil))
+  end
+
+  def test_parse_offset_hour_minute_second
+    t = Time.at(-100000000000).utc
+    assert_equal(t, Time.parse("1200-02-15 BC 14:13:20-00"))
+    assert_equal(t, Time.parse("1200-02-15 BC 14:13:20-00:00"))
+    assert_equal(t, Time.parse("1200-02-15 BC 14:13:20-00:00:00"))
+  end
+
   def test_parse_leap_second
     t = Time.utc(1998,12,31,23,59,59)
     assert_equal(t, Time.parse("Thu Dec 31 23:59:59 UTC 1998"))
@@ -448,6 +462,17 @@ class TestTimeExtension < Test::Unit::TestCase # :nodoc:
     assert_equal(true, t.utc?)
   end
 
+  def test_strptime_s_N
+    assert_equal(Time.at(1, 500000), Time.strptime("1.5", "%s.%N"))
+    assert_equal(Time.at(-2, 500000), Time.strptime("-1.5", "%s.%N"))
+    t = Time.strptime("1.000000000001", "%s.%N")
+    assert_equal(1, t.to_i)
+    assert_equal(Rational("0.000000000001"), t.subsec)
+    t = Time.strptime("-1.000000000001", "%s.%N")
+    assert_equal(-2, t.to_i)
+    assert_equal(1-Rational("0.000000000001"), t.subsec)
+  end
+
   def test_strptime_Ymd_z
     t = Time.strptime('20010203 -0200', '%Y%m%d %z')
     assert_equal(2001, t.year)
