@@ -2,7 +2,7 @@
 
   vm_dump.c -
 
-  $Author: naruse $
+  $Author: stomar $
 
   Copyright (C) 2004-2007 Koichi Sasada
 
@@ -21,14 +21,15 @@
 
 #define MAX_POSBUF 128
 
-#define VM_CFP_CNT(th, cfp) \
-  ((rb_control_frame_t *)((th)->stack + (th)->stack_size) - (rb_control_frame_t *)(cfp))
+#define VM_CFP_CNT(ec, cfp) \
+  ((rb_control_frame_t *)((ec)->vm_stack + (ec)->vm_stack_size) - \
+   (rb_control_frame_t *)(cfp))
 
 static void
-control_frame_dump(rb_thread_t *th, rb_control_frame_t *cfp)
+control_frame_dump(const rb_execution_context_t *ec, const rb_control_frame_t *cfp)
 {
     ptrdiff_t pc = -1;
-    ptrdiff_t ep = cfp->ep - th->stack;
+    ptrdiff_t ep = cfp->ep - ec->vm_stack;
     char ep_in_heap = ' ';
     char posbuf[MAX_POSBUF+1];
     int line = 0;
@@ -36,11 +37,9 @@ control_frame_dump(rb_thread_t *th, rb_control_frame_t *cfp)
     const char *magic, *iseq_name = "-", *selfstr = "-", *biseq_name = "-";
     VALUE tmp;
 
-    if (cfp->block_iseq != 0 && BUILTIN_TYPE(cfp->block_iseq) != T_NODE) {
-	biseq_name = "";	/* RSTRING(cfp->block_iseq->location.label)->ptr; */
-    }
+    const rb_callable_method_entry_t *me;
 
-    if (ep < 0 || (size_t)ep > th->stack_size) {
+    if (ep < 0 || (size_t)ep > ec->vm_stack_size) {
 	ep = (ptrdiff_t)cfp->ep;
 	ep_in_heap = 'p';
     }
@@ -61,12 +60,6 @@ control_frame_dump(rb_thread_t *th, rb_control_frame_t *cfp)
       case VM_FRAME_MAGIC_CFUNC:
 	magic = "CFUNC";
 	break;
-      case VM_FRAME_MAGIC_PROC:
-	magic = "PROC";
-	break;
-      case VM_FRAME_MAGIC_LAMBDA:
-	magic = "LAMBDA";
-	break;
       case VM_FRAME_MAGIC_IFUNC:
 	magic = "IFUNC";
 	break;
@@ -93,39 +86,46 @@ control_frame_dump(rb_thread_t *th, rb_control_frame_t *cfp)
     }
 
     if (cfp->iseq != 0) {
+#define RUBY_VM_IFUNC_P(ptr) imemo_type_p((VALUE)ptr, imemo_ifunc)
 	if (RUBY_VM_IFUNC_P(cfp->iseq)) {
 	    iseq_name = "<ifunc>";
 	}
+	else if (SYMBOL_P(cfp->iseq)) {
+	    tmp = rb_sym2str((VALUE)cfp->iseq);
+	    iseq_name = RSTRING_PTR(tmp);
+	    snprintf(posbuf, MAX_POSBUF, ":%s", iseq_name);
+	    line = -1;
+	}
 	else {
-	    pc = cfp->pc - cfp->iseq->iseq_encoded;
-	    iseq_name = RSTRING_PTR(cfp->iseq->location.label);
+	    pc = cfp->pc - cfp->iseq->body->iseq_encoded;
+	    iseq_name = RSTRING_PTR(cfp->iseq->body->location.label);
 	    line = rb_vm_get_sourceline(cfp);
 	    if (line) {
-		snprintf(posbuf, MAX_POSBUF, "%s:%d", RSTRING_PTR(cfp->iseq->location.path), line);
+		snprintf(posbuf, MAX_POSBUF, "%s:%d", RSTRING_PTR(rb_iseq_path(cfp->iseq)), line);
 	    }
 	}
     }
-    else if (cfp->me) {
-	iseq_name = rb_id2name(cfp->me->def->original_id);
+    else if ((me = rb_vm_frame_method_entry(cfp)) != NULL) {
+	iseq_name = rb_id2name(me->def->original_id);
 	snprintf(posbuf, MAX_POSBUF, ":%s", iseq_name);
 	line = -1;
     }
 
     fprintf(stderr, "c:%04"PRIdPTRDIFF" ",
-	    ((rb_control_frame_t *)(th->stack + th->stack_size) - cfp));
+	    ((rb_control_frame_t *)(ec->vm_stack + ec->vm_stack_size) - cfp));
     if (pc == -1) {
 	fprintf(stderr, "p:---- ");
     }
     else {
 	fprintf(stderr, "p:%04"PRIdPTRDIFF" ", pc);
     }
-    fprintf(stderr, "s:%04"PRIdPTRDIFF" ", cfp->sp - th->stack);
+    fprintf(stderr, "s:%04"PRIdPTRDIFF" ", cfp->sp - ec->vm_stack);
     fprintf(stderr, ep_in_heap == ' ' ? "e:%06"PRIdPTRDIFF" " : "E:%06"PRIxPTRDIFF" ", ep % 10000);
     fprintf(stderr, "%-6s", magic);
     if (line) {
 	fprintf(stderr, " %s", posbuf);
     }
-    if (VM_FRAME_TYPE_FINISH_P(cfp)) {
+    if (VM_FRAME_FINISHED_P(cfp)) {
 	fprintf(stderr, " [FINISH]");
     }
     if (0) {
@@ -138,19 +138,19 @@ control_frame_dump(rb_thread_t *th, rb_control_frame_t *cfp)
 }
 
 void
-rb_vmdebug_stack_dump_raw(rb_thread_t *th, rb_control_frame_t *cfp)
+rb_vmdebug_stack_dump_raw(const rb_execution_context_t *ec, const rb_control_frame_t *cfp)
 {
 #if 0
     VALUE *sp = cfp->sp, *ep = cfp->ep;
     VALUE *p, *st, *t;
 
     fprintf(stderr, "-- stack frame ------------\n");
-    for (p = st = th->stack; p < sp; p++) {
+    for (p = st = ec->vm_stack; p < sp; p++) {
 	fprintf(stderr, "%04ld (%p): %08"PRIxVALUE, (long)(p - st), p, *p);
 
 	t = (VALUE *)*p;
-	if (th->stack <= t && t < sp) {
-	    fprintf(stderr, " (= %ld)", (long)((VALUE *)GC_GUARDED_PTR_REF(t) - th->stack));
+	if (ec->vm_stack <= t && t < sp) {
+	    fprintf(stderr, " (= %ld)", (long)((VALUE *)GC_GUARDED_PTR_REF(t) - ec->vm_stack));
 	}
 
 	if (p == ep)
@@ -162,8 +162,8 @@ rb_vmdebug_stack_dump_raw(rb_thread_t *th, rb_control_frame_t *cfp)
 
     fprintf(stderr, "-- Control frame information "
 	    "-----------------------------------------------\n");
-    while ((void *)cfp < (void *)(th->stack + th->stack_size)) {
-	control_frame_dump(th, cfp);
+    while ((void *)cfp < (void *)(ec->vm_stack + ec->vm_stack_size)) {
+	control_frame_dump(ec, cfp);
 	cfp++;
     }
     fprintf(stderr, "\n");
@@ -172,32 +172,25 @@ rb_vmdebug_stack_dump_raw(rb_thread_t *th, rb_control_frame_t *cfp)
 void
 rb_vmdebug_stack_dump_raw_current(void)
 {
-    rb_thread_t *th = GET_THREAD();
-    rb_vmdebug_stack_dump_raw(th, th->cfp);
+    const rb_execution_context_t *ec = GET_EC();
+    rb_vmdebug_stack_dump_raw(ec, ec->cfp);
 }
 
 void
-rb_vmdebug_env_dump_raw(rb_env_t *env, VALUE *ep)
+rb_vmdebug_env_dump_raw(const rb_env_t *env, const VALUE *ep)
 {
-    int i;
+    unsigned int i;
     fprintf(stderr, "-- env --------------------\n");
 
     while (env) {
 	fprintf(stderr, "--\n");
 	for (i = 0; i < env->env_size; i++) {
-	    fprintf(stderr, "%04d: %08"PRIxVALUE" (%p)", -env->local_size + i, env->env[i],
-		   (void *)&env->env[i]);
-	    if (&env->env[i] == ep)
-		fprintf(stderr, " <- ep");
+	    fprintf(stderr, "%04d: %08"PRIxVALUE" (%p)", i, env->env[i], (void *)&env->env[i]);
+	    if (&env->env[i] == ep) fprintf(stderr, " <- ep");
 	    fprintf(stderr, "\n");
 	}
 
-	if (env->prev_envval != 0) {
-	    GetEnvPtr(env->prev_envval, env);
-	}
-	else {
-	    env = 0;
-	}
+	env = rb_vm_env_prev_env(env);
     }
     fprintf(stderr, "---------------------------\n");
 }
@@ -205,68 +198,51 @@ rb_vmdebug_env_dump_raw(rb_env_t *env, VALUE *ep)
 void
 rb_vmdebug_proc_dump_raw(rb_proc_t *proc)
 {
-    rb_env_t *env;
+    const rb_env_t *env;
     char *selfstr;
-    VALUE val = rb_inspect(proc->block.self);
+    VALUE val = rb_inspect(vm_block_self(&proc->block));
     selfstr = StringValueCStr(val);
 
     fprintf(stderr, "-- proc -------------------\n");
     fprintf(stderr, "self: %s\n", selfstr);
-    GetEnvPtr(proc->envval, env);
-    rb_vmdebug_env_dump_raw(env, proc->block.ep);
+    env = VM_ENV_ENVVAL_PTR(vm_block_ep(&proc->block));
+    rb_vmdebug_env_dump_raw(env, vm_block_ep(&proc->block));
 }
 
 void
 rb_vmdebug_stack_dump_th(VALUE thval)
 {
-    rb_thread_t *th;
-    GetThreadPtr(thval, th);
-    rb_vmdebug_stack_dump_raw(th, th->cfp);
+    rb_thread_t *target_th = rb_thread_ptr(thval);
+    rb_vmdebug_stack_dump_raw(target_th->ec, target_th->ec->cfp);
 }
 
 #if VMDEBUG > 2
 
 /* copy from vm.c */
-static VALUE *
-vm_base_ptr(rb_control_frame_t *cfp)
+static const VALUE *
+vm_base_ptr(const rb_control_frame_t *cfp)
 {
-    rb_control_frame_t *prev_cfp = RUBY_VM_PREVIOUS_CONTROL_FRAME(cfp);
-    VALUE *bp = prev_cfp->sp + cfp->iseq->local_size + 1;
+    const rb_control_frame_t *prev_cfp = RUBY_VM_PREVIOUS_CONTROL_FRAME(cfp);
+    const VALUE *bp = prev_cfp->sp + cfp->iseq->body->local_table_size + VM_ENV_DATA_SIZE;
 
-    if (cfp->iseq->type == ISEQ_TYPE_METHOD) {
+    if (cfp->iseq->body->type == ISEQ_TYPE_METHOD) {
 	bp += 1;
     }
     return bp;
 }
 
 static void
-vm_stack_dump_each(rb_thread_t *th, rb_control_frame_t *cfp)
+vm_stack_dump_each(const rb_execution_context_t *ec, const rb_control_frame_t *cfp)
 {
-    int i;
-
+    int i, argc = 0, local_table_size = 0;
     VALUE rstr;
     VALUE *sp = cfp->sp;
-    VALUE *ep = cfp->ep;
+    const VALUE *ep = cfp->ep;
 
-    int argc = 0, local_size = 0;
-    const char *name;
-    rb_iseq_t *iseq = cfp->iseq;
-
-    if (iseq == 0) {
-	if (RUBYVM_CFUNC_FRAME_P(cfp)) {
-	    name = rb_id2name(cfp->me->called_id);
-	}
-	else {
-	    name = "?";
-	}
-    }
-    else if (RUBY_VM_IFUNC_P(iseq)) {
-	name = "<ifunc>";
-    }
-    else {
-	argc = iseq->param.lead_num;
-	local_size = iseq->local_size;
-	name = RSTRING_PTR(iseq->location.label);
+    if (VM_FRAME_RUBYFRAME_P(cfp)) {
+	const rb_iseq_t *iseq = cfp->iseq;
+	argc = iseq->body->param.lead_num;
+	local_table_size = iseq->body->local_table_size;
     }
 
     /* stack trace header */
@@ -275,24 +251,21 @@ vm_stack_dump_each(rb_thread_t *th, rb_control_frame_t *cfp)
 	VM_FRAME_TYPE(cfp) == VM_FRAME_MAGIC_TOP   ||
 	VM_FRAME_TYPE(cfp) == VM_FRAME_MAGIC_BLOCK ||
 	VM_FRAME_TYPE(cfp) == VM_FRAME_MAGIC_CLASS ||
-	VM_FRAME_TYPE(cfp) == VM_FRAME_MAGIC_PROC  ||
-	VM_FRAME_TYPE(cfp) == VM_FRAME_MAGIC_LAMBDA||
 	VM_FRAME_TYPE(cfp) == VM_FRAME_MAGIC_CFUNC ||
 	VM_FRAME_TYPE(cfp) == VM_FRAME_MAGIC_IFUNC ||
 	VM_FRAME_TYPE(cfp) == VM_FRAME_MAGIC_EVAL  ||
 	VM_FRAME_TYPE(cfp) == VM_FRAME_MAGIC_RESCUE)
-      {
+    {
+	const VALUE *ptr = ep - local_table_size;
 
-	VALUE *ptr = ep - local_size;
-
-	control_frame_dump(th, cfp);
+	control_frame_dump(ec, cfp);
 
 	for (i = 0; i < argc; i++) {
 	    rstr = rb_inspect(*ptr);
 	    fprintf(stderr, "  arg   %2d: %8s (%p)\n", i, StringValueCStr(rstr),
 		   (void *)ptr++);
 	}
-	for (; i < local_size - 1; i++) {
+	for (; i < local_table_size - 1; i++) {
 	    rstr = rb_inspect(*ptr);
 	    fprintf(stderr, "  local %2d: %8s (%p)\n", i, StringValueCStr(rstr),
 		   (void *)ptr++);
@@ -300,19 +273,24 @@ vm_stack_dump_each(rb_thread_t *th, rb_control_frame_t *cfp)
 
 	ptr = vm_base_ptr(cfp);
 	for (; ptr < sp; ptr++, i++) {
-	    if (*ptr == Qundef) {
+	    switch (TYPE(*ptr)) {
+	      case T_UNDEF:
 		rstr = rb_str_new2("undef");
-	    }
-	    else {
+		break;
+	      case T_IMEMO:
+		rstr = rb_str_new2("imemo"); /* TODO: can put mode detail information */
+		break;
+	      default:
 		rstr = rb_inspect(*ptr);
+		break;
 	    }
 	    fprintf(stderr, "  stack %2d: %8s (%"PRIdPTRDIFF")\n", i, StringValueCStr(rstr),
-		    (ptr - th->stack));
+		    (ptr - ec->vm_stack));
 	}
     }
-    else if (VM_FRAME_TYPE_FINISH_P(cfp)) {
-	if ((th)->stack + (th)->stack_size > (VALUE *)(cfp + 1)) {
-	    vm_stack_dump_each(th, cfp + 1);
+    else if (VM_FRAME_FINISHED_P(cfp)) {
+	if (ec->vm_stack + ec->vm_stack_size > (VALUE *)(cfp + 1)) {
+	    vm_stack_dump_each(ec, cfp + 1);
 	}
 	else {
 	    /* SDR(); */
@@ -325,52 +303,50 @@ vm_stack_dump_each(rb_thread_t *th, rb_control_frame_t *cfp)
 #endif
 
 void
-rb_vmdebug_debug_print_register(rb_thread_t *th)
+rb_vmdebug_debug_print_register(const rb_execution_context_t *ec)
 {
-    rb_control_frame_t *cfp = th->cfp;
+    rb_control_frame_t *cfp = ec->cfp;
     ptrdiff_t pc = -1;
-    ptrdiff_t ep = cfp->ep - th->stack;
+    ptrdiff_t ep = cfp->ep - ec->vm_stack;
     ptrdiff_t cfpi;
 
-    if (RUBY_VM_NORMAL_ISEQ_P(cfp->iseq)) {
-	pc = cfp->pc - cfp->iseq->iseq_encoded;
+    if (VM_FRAME_RUBYFRAME_P(cfp)) {
+	pc = cfp->pc - cfp->iseq->body->iseq_encoded;
     }
 
-    if (ep < 0 || (size_t)ep > th->stack_size) {
+    if (ep < 0 || (size_t)ep > ec->vm_stack_size) {
 	ep = -1;
     }
 
-    cfpi = ((rb_control_frame_t *)(th->stack + th->stack_size)) - cfp;
+    cfpi = ((rb_control_frame_t *)(ec->vm_stack + ec->vm_stack_size)) - cfp;
     fprintf(stderr, "  [PC] %04"PRIdPTRDIFF", [SP] %04"PRIdPTRDIFF", [EP] %04"PRIdPTRDIFF", [CFP] %04"PRIdPTRDIFF"\n",
-	    pc, (cfp->sp - th->stack), ep, cfpi);
+	    pc, (cfp->sp - ec->vm_stack), ep, cfpi);
 }
 
 void
 rb_vmdebug_thread_dump_regs(VALUE thval)
 {
-    rb_thread_t *th;
-    GetThreadPtr(thval, th);
-    rb_vmdebug_debug_print_register(th);
+    rb_vmdebug_debug_print_register(rb_thread_ptr(thval)->ec);
 }
 
 void
-rb_vmdebug_debug_print_pre(rb_thread_t *th, rb_control_frame_t *cfp,VALUE *_pc)
+rb_vmdebug_debug_print_pre(const rb_execution_context_t *ec, const rb_control_frame_t *cfp, const VALUE *_pc)
 {
-    rb_iseq_t *iseq = cfp->iseq;
+    const rb_iseq_t *iseq = cfp->iseq;
 
     if (iseq != 0) {
-	ptrdiff_t pc = _pc - iseq->iseq_encoded;
+	ptrdiff_t pc = _pc - iseq->body->iseq_encoded;
 	int i;
 
-	for (i=0; i<(int)VM_CFP_CNT(th, cfp); i++) {
+	for (i=0; i<(int)VM_CFP_CNT(ec, cfp); i++) {
 	    printf(" ");
 	}
 	printf("| ");
-	if(0)printf("[%03ld] ", (long)(cfp->sp - th->stack));
+	if(0)printf("[%03ld] ", (long)(cfp->sp - ec->vm_stack));
 
-	/* printf("%3"PRIdPTRDIFF" ", VM_CFP_CNT(th, cfp)); */
+	/* printf("%3"PRIdPTRDIFF" ", VM_CFP_CNT(ec, cfp)); */
 	if (pc >= 0) {
-	    const VALUE *iseq_original = rb_iseq_original_iseq(iseq);
+	    const VALUE *iseq_original = rb_iseq_original_iseq((rb_iseq_t *)iseq);
 
 	    rb_iseq_disasm_insn(0, iseq_original, (size_t)pc, iseq, 0);
 	}
@@ -378,12 +354,12 @@ rb_vmdebug_debug_print_pre(rb_thread_t *th, rb_control_frame_t *cfp,VALUE *_pc)
 
 #if VMDEBUG > 3
     fprintf(stderr, "        (1)");
-    rb_vmdebug_debug_print_register(th);
+    rb_vmdebug_debug_print_register(ec);
 #endif
 }
 
 void
-rb_vmdebug_debug_print_post(rb_thread_t *th, rb_control_frame_t *cfp
+rb_vmdebug_debug_print_post(const rb_execution_context_t *ec, const rb_control_frame_t *cfp
 #if OPT_STACK_CACHING
 		 , VALUE reg_a, VALUE reg_b
 #endif
@@ -395,13 +371,13 @@ rb_vmdebug_debug_print_post(rb_thread_t *th, rb_control_frame_t *cfp
 
 #if VMDEBUG > 3
     fprintf(stderr, "        (2)");
-    rb_vmdebug_debug_print_register(th);
+    rb_vmdebug_debug_print_register(ec);
 #endif
-    /* stack_dump_raw(th, cfp); */
+    /* stack_dump_raw(ec, cfp); */
 
 #if VMDEBUG > 2
-    /* stack_dump_thobj(th); */
-    vm_stack_dump_each(th, th->cfp);
+    /* stack_dump_thobj(ec); */
+    vm_stack_dump_each(ec, ec->cfp);
 
 #if OPT_STACK_CACHING
     {
@@ -420,10 +396,8 @@ rb_vmdebug_debug_print_post(rb_thread_t *th, rb_control_frame_t *cfp
 VALUE
 rb_vmdebug_thread_dump_state(VALUE self)
 {
-    rb_thread_t *th;
-    rb_control_frame_t *cfp;
-    GetThreadPtr(self, th);
-    cfp = th->cfp;
+    rb_thread_t *th = rb_thread_ptr(self);
+    rb_control_frame_t *cfp = th->ec->cfp;
 
     fprintf(stderr, "Thread state dump:\n");
     fprintf(stderr, "pc : %p, sp : %p\n", (void *)cfp->pc, (void *)cfp->sp);
@@ -436,9 +410,10 @@ rb_vmdebug_thread_dump_state(VALUE self)
 # ifdef HAVE_LIBUNWIND
 #  undef backtrace
 #  define backtrace unw_backtrace
-# elif defined(__APPLE__) && defined(__x86_64__)
+# elif defined(__APPLE__) && defined(__x86_64__) && defined(HAVE_LIBUNWIND_H)
 #  define UNW_LOCAL_ONLY
 #  include <libunwind.h>
+#  include <sys/mman.h>
 #  undef backtrace
 int
 backtrace(void **trace, int size)
@@ -465,7 +440,9 @@ darwin_sigtramp:
     /* darwin's bundled libunwind doesn't support signal trampoline */
     {
 	ucontext_t *uctx;
-	/* get _sigtramp's ucontext_t and set values to cursor
+	char vec[1];
+	int r;
+	/* get previous frame information from %rbx at _sigtramp and set values to cursor
 	 * http://www.opensource.apple.com/source/Libc/Libc-825.25/i386/sys/_sigtramp.s
 	 * http://www.opensource.apple.com/source/libunwind/libunwind-35.1/src/unw_getcontext.s
 	 */
@@ -487,10 +464,43 @@ darwin_sigtramp:
 	unw_set_reg(&cursor, UNW_X86_64_R13, uctx->uc_mcontext->__ss.__r13);
 	unw_set_reg(&cursor, UNW_X86_64_R14, uctx->uc_mcontext->__ss.__r14);
 	unw_set_reg(&cursor, UNW_X86_64_R15, uctx->uc_mcontext->__ss.__r15);
-	ip = *(unw_word_t*)uctx->uc_mcontext->__ss.__rsp;
-	unw_set_reg(&cursor, UNW_REG_IP, ip);
-	trace[n++] = (void *)uctx->uc_mcontext->__ss.__rip;
+	ip = uctx->uc_mcontext->__ss.__rip;
+
+	/* There are 4 cases for SEGV:
+	 * (1) called invalid address
+	 * (2) read or write invalid address
+	 * (3) received signal
+	 *
+	 * Detail:
+	 * (1) called invalid address
+	 * In this case, saved ip is invalid address.
+	 * It needs to just save the address for the information,
+	 * skip the frame, and restore the frame calling the
+	 * invalid address from %rsp.
+	 * The problem is how to check whether the ip is valid or not.
+	 * This code uses mincore(2) and assume the address's page is
+	 * incore/referenced or not reflects the problem.
+	 * Note that High Sierra's mincore(2) may return -128.
+	 * (2) read or write invalid address
+	 * saved ip is valid. just restart backtracing.
+	 * (3) received signal in user space
+	 * Same as (2).
+	 * (4) received signal in kernel
+	 * In this case saved ip points just after syscall, but registers are
+	 * already overwritten by kernel. To fix register consistency,
+	 * skip libc's kernel wrapper.
+	 * To detect this case, just previous two bytes of ip is "\x0f\x05",
+	 * syscall instruction of x86_64.
+	 */
+	r = mincore((const void *)ip, 1, vec);
+	if (r || vec[0] <= 0 || memcmp((const char *)ip-2, "\x0f\x05", 2) == 0) {
+	    /* if segv is caused by invalid call or signal received in syscall */
+	    /* the frame is invalid; skip */
+	    trace[n++] = (void *)ip;
+	    ip = *(unw_word_t*)uctx->uc_mcontext->__ss.__rsp;
+	}
 	trace[n++] = (void *)ip;
+	unw_set_reg(&cursor, UNW_REG_IP, ip);
     }
     while (unw_step(&cursor) > 0) {
 	unw_get_reg(&cursor, UNW_REG_IP, &ip);
@@ -690,8 +700,8 @@ rb_print_backtrace(void)
 #if HAVE_BACKTRACE
 #define MAX_NATIVE_TRACE 1024
     static void *trace[MAX_NATIVE_TRACE];
-    int n = backtrace(trace, MAX_NATIVE_TRACE);
-#if defined(USE_ELF) && defined(HAVE_DLADDR)
+    int n = (int)backtrace(trace, MAX_NATIVE_TRACE);
+#if defined(USE_ELF) && defined(HAVE_DLADDR) && !defined(__sparc)
     rb_dump_backtrace_with_lines(n, trace);
 #else
     char **syms = backtrace_symbols(trace, n);
@@ -735,7 +745,11 @@ procstat_vm(struct procstat *procstat, struct kinfo_proc *kipp)
 		ptrwidth, "START", ptrwidth, "END", "PRT", "RES",
 		"PRES", "REF", "SHD", "FL", "TP", "PATH");
 
+#ifdef HAVE_PROCSTAT_GETVMMAP
 	freep = procstat_getvmmap(procstat, kipp, &cnt);
+#else
+	freep = kinfo_getvmmap(kipp->ki_pid, &cnt);
+#endif
 	if (freep == NULL)
 		return;
 	for (i = 0; i < cnt; i++) {
@@ -945,18 +959,6 @@ rb_vm_bugreport(const void *ctx)
 #endif
     const rb_vm_t *const vm = GET_VM();
 
-#if defined __APPLE__
-    fputs("-- Crash Report log information "
-	  "--------------------------------------------\n"
-	  "   See Crash Report log file under the one of following:\n"
-	  "     * ~/Library/Logs/CrashReporter\n"
-	  "     * /Library/Logs/CrashReporter\n"
-	  "     * ~/Library/Logs/DiagnosticReports\n"
-	  "     * /Library/Logs/DiagnosticReports\n"
-	  "   for more details.\n"
-	  "\n",
-	  stderr);
-#endif
     if (vm) {
 	SDR();
 	rb_backtrace_print_as_bugreport();
@@ -1000,12 +1002,21 @@ rb_vm_bugreport(const void *ctx)
 	    else if (RB_TYPE_P(name, T_CLASS) || RB_TYPE_P(name, T_MODULE)) {
 		const char *const type = RB_TYPE_P(name, T_CLASS) ?
 		    "class" : "module";
-		name = rb_class_name(name);
+		name = rb_search_class_path(rb_class_real(name));
+		if (!RB_TYPE_P(name, T_STRING)) {
+		    fprintf(stderr, " %4d %s:<unnamed>\n", i, type);
+		    continue;
+		}
 		fprintf(stderr, " %4d %s:%.*s\n", i, type,
 			LIMITED_NAME_LENGTH(name), RSTRING_PTR(name));
 	    }
 	    else {
-		VALUE klass = rb_class_name(CLASS_OF(name));
+		VALUE klass = rb_search_class_path(rb_obj_class(name));
+		if (!RB_TYPE_P(klass, T_STRING)) {
+		    fprintf(stderr, " %4d #<%p:%p>\n", i,
+			    (void *)CLASS_OF(name), (void *)name);
+		    continue;
+		}
 		fprintf(stderr, " %4d #<%.*s:%p>\n", i,
 			LIMITED_NAME_LENGTH(klass), RSTRING_PTR(klass),
 			(void *)name);
@@ -1055,3 +1066,25 @@ rb_vm_bugreport(const void *ctx)
 #endif /* __FreeBSD__ */
     }
 }
+
+#ifdef NON_SCALAR_THREAD_ID
+const char *ruby_fill_thread_id_string(rb_nativethread_id_t thid, rb_thread_id_string_t buf);
+#endif
+
+void
+rb_vmdebug_stack_dump_all_threads(void)
+{
+    rb_vm_t *vm = GET_VM();
+    rb_thread_t *th = NULL;
+
+    list_for_each(&vm->living_threads, th, vmlt_node) {
+#ifdef NON_SCALAR_THREAD_ID
+	rb_thread_id_string_t buf;
+	ruby_fill_thread_id_string(th->thread_id, buf);
+	fprintf(stderr, "th: %p, native_id: %s\n", th, buf);
+#else
+	fprintf(stderr, "th: %p, native_id: %p\n", th, (void *)th->thread_id);
+#endif
+	rb_vmdebug_stack_dump_raw(th->ec, th->ec->cfp);
+    }
+}
